# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: engine.proto
# Protobuf Python Version: 6.31.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    31,
    1,
    '',
    'engine.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0c\x65ngine.proto\x12\x06\x65ngine\"\xd1\x01\n\x0cOrderRequest\x12\x0e\n\x06symbol\x18\x01 \x01(\t\x12\x0c\n\x04side\x18\x02 \x01(\t\x12\x12\n\norder_type\x18\x03 \x01(\t\x12\x10\n\x08quantity\x18\x04 \x01(\t\x12\x13\n\x0blimit_price\x18\x05 \x01(\t\x12\x12\n\nstop_price\x18\x06 \x01(\t\x12\x15\n\rtime_in_force\x18\x07 \x01(\t\x12\x13\n\x0bstrategy_id\x18\x08 \x01(\t\x12\x17\n\x0f\x63lient_order_id\x18\t \x01(\t\x12\x0f\n\x07user_id\x18\n \x01(\t\"\xbc\x01\n\rOrderResponse\x12\x10\n\x08order_id\x18\x01 \x01(\t\x12\x17\n\x0f\x63lient_order_id\x18\x02 \x01(\t\x12\x0e\n\x06status\x18\x03 \x01(\t\x12\x12\n\nfilled_qty\x18\x04 \x01(\t\x12\x18\n\x10\x66illed_avg_price\x18\x05 \x01(\t\x12\x17\n\x0fsubmitted_at_ns\x18\x06 \x01(\x03\x12\x1a\n\x12\x61\x63knowledged_at_ns\x18\x07 \x01(\x03\x12\r\n\x05\x65rror\x18\x08 \x01(\t\"!\n\rCancelRequest\x12\x10\n\x08order_id\x18\x01 \x01(\t\"0\n\x0e\x43\x61ncelResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\r\n\x05\x65rror\x18\x02 \x01(\t\"\x12\n\x10PositionsRequest\"\x81\x01\n\x08Position\x12\x0e\n\x06symbol\x18\x01 \x01(\t\x12\x10\n\x08quantity\x18\x02 \x01(\t\x12\x17\n\x0f\x61vg_entry_price\x18\x03 \x01(\t\x12\x14\n\x0cmarket_value\x18\x04 \x01(\t\x12\x16\n\x0eunrealized_pnl\x18\x05 \x01(\t\x12\x0c\n\x04side\x18\x06 \x01(\t\"8\n\x11PositionsResponse\x12#\n\tpositions\x18\x01 \x03(\x0b\x32\x10.engine.Position\"\x0f\n\rStatusRequest\"\xa1\x02\n\x0eStatusResponse\x12\x13\n\x0b\x65ngine_name\x18\x01 \x01(\t\x12\x0c\n\x04mode\x18\x02 \x01(\t\x12\x16\n\x0euptime_seconds\x18\x03 \x01(\t\x12\x15\n\rorders_placed\x18\x04 \x01(\x03\x12\x15\n\rorders_filled\x18\x05 \x01(\x03\x12\x1c\n\x14market_data_messages\x18\x06 \x01(\x03\x12\x1c\n\x14\x61vg_tick_to_order_us\x18\x07 \x01(\x01\x12\x1c\n\x14p99_tick_to_order_us\x18\x08 \x01(\x01\x12\x1a\n\x12subscribed_symbols\x18\t \x03(\t\x12\x16\n\x0egateway_status\x18\n \x01(\t\x12\x18\n\x10\x61\x63tive_executors\x18\x0b \x01(\x05\"\x97\x01\n\x11\x43redentialsConfig\x12\x0e\n\x06\x62roker\x18\x01 \x01(\t\x12\x0f\n\x07\x61pi_key\x18\x02 \x01(\t\x12\x12\n\napi_secret\x18\x03 \x01(\t\x12\x13\n\x0b\x65nvironment\x18\x04 \x01(\t\x12\x12\n\naccount_id\x18\x05 \x01(\t\x12\x0f\n\x07user_id\x18\x06 \x01(\t\x12\x13\n\x0bstrategy_id\x18\x07 \x01(\t\"\x10\n\x0e\x41\x63\x63ountRequest\"Y\n\x0f\x41\x63\x63ountResponse\x12\x12\n\naccount_id\x18\x01 \x01(\t\x12\x14\n\x0c\x62uying_power\x18\x02 \x01(\t\x12\x0e\n\x06\x65quity\x18\x03 \x01(\t\x12\x0c\n\x04\x63\x61sh\x18\x04 \x01(\t\"#\n\x0fGetOrderRequest\x12\x10\n\x08order_id\x18\x01 \x01(\t\"j\n\x0c\x43onfigUpdate\x12\x0f\n\x07symbols\x18\x01 \x03(\t\x12\x1d\n\x15max_position_size_usd\x18\x02 \x01(\t\x12\x1c\n\x14max_loss_per_day_usd\x18\x03 \x01(\t\x12\x0c\n\x04mode\x18\x04 \x01(\t\"0\n\x0e\x43onfigResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\r\n\x05\x65rror\x18\x02 \x01(\t\"7\n\x10SubscribeRequest\x12\x0f\n\x07symbols\x18\x01 \x03(\t\x12\x12\n\ndata_types\x18\x02 \x03(\t\"\xa6\x01\n\x10MarketDataUpdate\x12\x0e\n\x06symbol\x18\x01 \x01(\t\x12\x11\n\tdata_type\x18\x02 \x01(\t\x12\r\n\x05price\x18\x03 \x01(\x01\x12\x0c\n\x04size\x18\x04 \x01(\x01\x12\x0b\n\x03\x62id\x18\x05 \x01(\x01\x12\x0b\n\x03\x61sk\x18\x06 \x01(\x01\x12\x10\n\x08\x62id_size\x18\x07 \x01(\x01\x12\x10\n\x08\x61sk_size\x18\x08 \x01(\x01\x12\x14\n\x0ctimestamp_ns\x18\t \x01(\x03\"\x17\n\x15OrderSubscribeRequest\"|\n\x0bOrderUpdate\x12\x10\n\x08order_id\x18\x01 \x01(\t\x12\x0e\n\x06symbol\x18\x02 \x01(\t\x12\x0e\n\x06status\x18\x03 \x01(\t\x12\x12\n\nfilled_qty\x18\x04 \x01(\t\x12\x11\n\tavg_price\x18\x05 \x01(\t\x12\x14\n\x0ctimestamp_ns\x18\x06 \x01(\x03\"\x9b\x01\n\x0eStrategyConfig\x12\x15\n\rstrategy_type\x18\x01 \x01(\t\x12\x32\n\x06params\x18\x02 \x03(\x0b\x32\".engine.StrategyConfig.ParamsEntry\x12\x0f\n\x07symbols\x18\x03 \x03(\t\x1a-\n\x0bParamsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\x90\x01\n\nRiskConfig\x12\x1d\n\x15max_position_size_usd\x18\x01 \x01(\x01\x12\"\n\x1amax_portfolio_exposure_usd\x18\x02 \x01(\x01\x12!\n\x19max_order_rate_per_second\x18\x03 \x01(\r\x12\x1c\n\x14max_loss_per_day_usd\x18\x04 \x01(\x01\"\xca\x01\n\x15\x44\x65ployStrategyRequest\x12\x0f\n\x07user_id\x18\x01 \x01(\t\x12\x13\n\x0bstrategy_id\x18\x02 \x01(\t\x12.\n\x0b\x63redentials\x18\x03 \x01(\x0b\x32\x19.engine.CredentialsConfig\x12(\n\x08strategy\x18\x04 \x01(\x0b\x32\x16.engine.StrategyConfig\x12 \n\x04risk\x18\x05 \x01(\x0b\x32\x12.engine.RiskConfig\x12\x0f\n\x07symbols\x18\x06 \x03(\t\"M\n\x16\x44\x65ployStrategyResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\x13\n\x0b\x65xecutor_id\x18\x02 \x01(\t\x12\r\n\x05\x65rror\x18\x03 \x01(\t\";\n\x13StopStrategyRequest\x12\x0f\n\x07user_id\x18\x01 \x01(\t\x12\x13\n\x0bstrategy_id\x18\x02 \x01(\t\"\xa4\x01\n\x0c\x45xecutorInfo\x12\x0f\n\x07user_id\x18\x01 \x01(\t\x12\x13\n\x0bstrategy_id\x18\x02 \x01(\t\x12\x0e\n\x06status\x18\x03 \x01(\t\x12\x15\n\rorders_placed\x18\x04 \x01(\x03\x12\x15\n\rorders_filled\x18\x05 \x01(\x03\x12\x1f\n\x17market_events_processed\x18\x06 \x01(\x03\x12\x0f\n\x07symbols\x18\x07 \x03(\t\"\x16\n\x14ListExecutorsRequest\"@\n\x15ListExecutorsResponse\x12\'\n\texecutors\x18\x01 \x03(\x0b\x32\x14.engine.ExecutorInfo\"+\n\x18GetExecutorStatusRequest\x12\x0f\n\x07user_id\x18\x01 \x01(\t2\xda\x07\n\rEngineService\x12\x39\n\nPlaceOrder\x12\x14.engine.OrderRequest\x1a\x15.engine.OrderResponse\x12<\n\x0b\x43\x61ncelOrder\x12\x15.engine.CancelRequest\x1a\x16.engine.CancelResponse\x12\x43\n\x0cGetPositions\x12\x18.engine.PositionsRequest\x1a\x19.engine.PositionsResponse\x12=\n\nGetAccount\x12\x16.engine.AccountRequest\x1a\x17.engine.AccountResponse\x12:\n\x08GetOrder\x12\x17.engine.GetOrderRequest\x1a\x15.engine.OrderResponse\x12@\n\x0fGetEngineStatus\x12\x15.engine.StatusRequest\x1a\x16.engine.StatusResponse\x12I\n\x14\x43onfigureCredentials\x12\x19.engine.CredentialsConfig\x1a\x16.engine.ConfigResponse\x12<\n\x0cUpdateConfig\x12\x14.engine.ConfigUpdate\x1a\x16.engine.ConfigResponse\x12H\n\x10StreamMarketData\x12\x18.engine.SubscribeRequest\x1a\x18.engine.MarketDataUpdate0\x01\x12J\n\x12StreamOrderUpdates\x12\x1d.engine.OrderSubscribeRequest\x1a\x13.engine.OrderUpdate0\x01\x12O\n\x0e\x44\x65ployStrategy\x12\x1d.engine.DeployStrategyRequest\x1a\x1e.engine.DeployStrategyResponse\x12\x43\n\x0cStopStrategy\x12\x1b.engine.StopStrategyRequest\x1a\x16.engine.ConfigResponse\x12L\n\rListExecutors\x12\x1c.engine.ListExecutorsRequest\x1a\x1d.engine.ListExecutorsResponse\x12K\n\x11GetExecutorStatus\x12 .engine.GetExecutorStatusRequest\x1a\x14.engine.ExecutorInfob\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'engine_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_STRATEGYCONFIG_PARAMSENTRY']._loaded_options = None
  _globals['_STRATEGYCONFIG_PARAMSENTRY']._serialized_options = b'8\001'
  _globals['_ORDERREQUEST']._serialized_start=25
  _globals['_ORDERREQUEST']._serialized_end=234
  _globals['_ORDERRESPONSE']._serialized_start=237
  _globals['_ORDERRESPONSE']._serialized_end=425
  _globals['_CANCELREQUEST']._serialized_start=427
  _globals['_CANCELREQUEST']._serialized_end=460
  _globals['_CANCELRESPONSE']._serialized_start=462
  _globals['_CANCELRESPONSE']._serialized_end=510
  _globals['_POSITIONSREQUEST']._serialized_start=512
  _globals['_POSITIONSREQUEST']._serialized_end=530
  _globals['_POSITION']._serialized_start=533
  _globals['_POSITION']._serialized_end=662
  _globals['_POSITIONSRESPONSE']._serialized_start=664
  _globals['_POSITIONSRESPONSE']._serialized_end=720
  _globals['_STATUSREQUEST']._serialized_start=722
  _globals['_STATUSREQUEST']._serialized_end=737
  _globals['_STATUSRESPONSE']._serialized_start=740
  _globals['_STATUSRESPONSE']._serialized_end=1029
  _globals['_CREDENTIALSCONFIG']._serialized_start=1032
  _globals['_CREDENTIALSCONFIG']._serialized_end=1183
  _globals['_ACCOUNTREQUEST']._serialized_start=1185
  _globals['_ACCOUNTREQUEST']._serialized_end=1201
  _globals['_ACCOUNTRESPONSE']._serialized_start=1203
  _globals['_ACCOUNTRESPONSE']._serialized_end=1292
  _globals['_GETORDERREQUEST']._serialized_start=1294
  _globals['_GETORDERREQUEST']._serialized_end=1329
  _globals['_CONFIGUPDATE']._serialized_start=1331
  _globals['_CONFIGUPDATE']._serialized_end=1437
  _globals['_CONFIGRESPONSE']._serialized_start=1439
  _globals['_CONFIGRESPONSE']._serialized_end=1487
  _globals['_SUBSCRIBEREQUEST']._serialized_start=1489
  _globals['_SUBSCRIBEREQUEST']._serialized_end=1544
  _globals['_MARKETDATAUPDATE']._serialized_start=1547
  _globals['_MARKETDATAUPDATE']._serialized_end=1713
  _globals['_ORDERSUBSCRIBEREQUEST']._serialized_start=1715
  _globals['_ORDERSUBSCRIBEREQUEST']._serialized_end=1738
  _globals['_ORDERUPDATE']._serialized_start=1740
  _globals['_ORDERUPDATE']._serialized_end=1864
  _globals['_STRATEGYCONFIG']._serialized_start=1867
  _globals['_STRATEGYCONFIG']._serialized_end=2022
  _globals['_STRATEGYCONFIG_PARAMSENTRY']._serialized_start=1977
  _globals['_STRATEGYCONFIG_PARAMSENTRY']._serialized_end=2022
  _globals['_RISKCONFIG']._serialized_start=2025
  _globals['_RISKCONFIG']._serialized_end=2169
  _globals['_DEPLOYSTRATEGYREQUEST']._serialized_start=2172
  _globals['_DEPLOYSTRATEGYREQUEST']._serialized_end=2374
  _globals['_DEPLOYSTRATEGYRESPONSE']._serialized_start=2376
  _globals['_DEPLOYSTRATEGYRESPONSE']._serialized_end=2453
  _globals['_STOPSTRATEGYREQUEST']._serialized_start=2455
  _globals['_STOPSTRATEGYREQUEST']._serialized_end=2514
  _globals['_EXECUTORINFO']._serialized_start=2517
  _globals['_EXECUTORINFO']._serialized_end=2681
  _globals['_LISTEXECUTORSREQUEST']._serialized_start=2683
  _globals['_LISTEXECUTORSREQUEST']._serialized_end=2705
  _globals['_LISTEXECUTORSRESPONSE']._serialized_start=2707
  _globals['_LISTEXECUTORSRESPONSE']._serialized_end=2771
  _globals['_GETEXECUTORSTATUSREQUEST']._serialized_start=2773
  _globals['_GETEXECUTORSTATUSREQUEST']._serialized_end=2816
  _globals['_ENGINESERVICE']._serialized_start=2819
  _globals['_ENGINESERVICE']._serialized_end=3805
# @@protoc_insertion_point(module_scope)



from __future__ import annotations

from .clients.async_ import AsyncCPZClient
from .clients.sync import CPZClient
from .common.cpz_ai import CPZAIClient
from .execution.enums import OrderSide, OrderType, TimeInForce
from .execution.models import (
    Account,
    Order,
    OrderReplaceRequest,
    OrderSubmitRequest,
    Position,
    Quote,
)
from .execution.router import BROKER_ALPACA, BROKER_HFT, BROKER_POLYMARKET, wait_for_order_recording

# Data layer exports
from .data.models import (
    Bar,
    Trade,
    News,
    OptionQuote,
    OptionContract,
    EconomicSeries,
    Filing,
    SocialPost,
    TimeFrame,
)
from .data.client import DataClient

# Risk analytics exports
from .risk.engine import RiskAnalytics
RiskEngine = RiskAnalytics  # backward compat
from .risk.models import (
    RiskSnapshot,
    StrategyRisk,
    MonteCarloResult,
    StressScenario,
    RollingMetrics,
    DrawdownAnalysis,
    PositionSizeResult,
    SharpeInference,
    FactorExposure,
    TailRiskMetrics,
)

# Portfolio decomposition (new in 2.3.0)
from cpz.risk import (
    euler_decomposition,
    mctr,
    optimize_portfolio,
    what_if_analysis,
    risk_budget_check,
    EulerDecomposition,
    MCTRResult,
    OptimizationResult,
    WhatIfResult,
    RiskBudgetResult,
    cointegration_test,
    cointegration_rolling,
    CointegrationResult,
    RollingCointegrationResult,
    custom_scenario,
    cascade_effects,
    historical_scenario,
    scenario_comparison,
    ScenarioImpact,
    CascadeResult,
    ScenarioComparison,
    fragility_score,
    correlation_clustering,
    regime_from_returns,
    FragilityResult,
    CorrelationClusterResult,
    RegimeResult,
)

# Simons AI Assistant exports
from .simons import SimonsClient, AsyncSimonsClient
from .simons_models import (
    SimonsResponse,
    SimonsChunk,
    SimonsMemory,
    TokenUsage,
    CodeProposal,
)

# ── NEW in 3.0.0 ─────────────────────────────────────────────────────

# Typed domain model
from .model.types import Price, Quantity
from .model.types import Money as TypedMoney
from .model.events import (
    Event,
    BarEvent,
    QuoteEvent,
    TradeEvent,
    OrderEvent,
    FillEvent,
    PositionEvent,
    TimeEvent,
    SignalEvent,
)

# Event bus
from .events.bus import EventBus

# Strategy framework
from .strategy.config import StrategyConfig
from .strategy.base import Strategy
from .strategy.runner import StrategyRunner
from .strategy.backtest import BacktestResult

# Pre-trade risk guard
from .execution.risk_guard import RiskGuard, RiskGuardViolation

# Execution algorithms
from .execution.algos import TWAPAlgorithm, VWAPAlgorithm, IcebergAlgorithm

__all__ = [
    # Clients
    "CPZClient",
    "AsyncCPZClient",
    "CPZAIClient",
    "DataClient",
    # Simons AI Assistant
    "SimonsClient",
    "AsyncSimonsClient",
    "SimonsResponse",
    "SimonsChunk",
    "SimonsMemory",
    "TokenUsage",
    "CodeProposal",
    # Execution
    "OrderSide",
    "OrderType",
    "TimeInForce",
    "OrderSubmitRequest",
    "OrderReplaceRequest",
    "Order",
    "Account",
    "Position",
    "Quote",
    "BROKER_ALPACA",
    "BROKER_HFT",
    "BROKER_POLYMARKET",
    "wait_for_order_recording",
    # Risk Analytics
    "RiskAnalytics",
    "RiskEngine",  # backward compat
    "RiskSnapshot",
    "StrategyRisk",
    "MonteCarloResult",
    "StressScenario",
    "RollingMetrics",
    "DrawdownAnalysis",
    "PositionSizeResult",
    "SharpeInference",
    "FactorExposure",
    "TailRiskMetrics",
    # Data Models
    "Bar",
    "Trade",
    "News",
    "OptionQuote",
    "OptionContract",
    "EconomicSeries",
    "Filing",
    "SocialPost",
    "TimeFrame",
    # ── NEW in 3.0.0 ──
    # Typed Domain Model
    "Price",
    "Quantity",
    "TypedMoney",
    "Event",
    "BarEvent",
    "QuoteEvent",
    "TradeEvent",
    "OrderEvent",
    "FillEvent",
    "PositionEvent",
    "TimeEvent",
    "SignalEvent",
    # Event Bus
    "EventBus",
    # Strategy Framework
    "Strategy",
    "StrategyConfig",
    "StrategyRunner",
    "BacktestResult",
    # Pre-Trade Risk Guard
    "RiskGuard",
    "RiskGuardViolation",
    # Execution Algorithms
    "TWAPAlgorithm",
    "VWAPAlgorithm",
    "IcebergAlgorithm",
]

__version__ = "3.0.0"



"""Base class for execution algorithms."""

from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional

from ..models import Order, OrderSubmitRequest

logger = logging.getLogger(__name__)


@dataclass
class AlgoSlice:
    """A child order slice produced by an execution algorithm."""

    seq: int
    symbol: str
    side: str
    qty: float
    order_type: str = "market"
    limit_price: Optional[float] = None
    scheduled_time: float = 0.0
    status: str = "pending"
    order_id: str = ""
    fill_price: float = 0.0
    fill_qty: float = 0.0


@dataclass
class AlgoState:
    """Running state of an execution algorithm."""

    parent_order: Optional[OrderSubmitRequest] = None
    total_qty: float = 0.0
    filled_qty: float = 0.0
    slices: List[AlgoSlice] = field(default_factory=list)
    status: str = "pending"
    avg_fill_price: float = 0.0
    total_cost: float = 0.0

    @property
    def remaining_qty(self) -> float:
        return self.total_qty - self.filled_qty

    @property
    def is_complete(self) -> bool:
        return self.filled_qty >= self.total_qty or self.status in ("completed", "canceled")

    @property
    def fill_pct(self) -> float:
        return (self.filled_qty / self.total_qty * 100) if self.total_qty > 0 else 0

    def record_fill(self, slice_seq: int, fill_qty: float, fill_price: float) -> None:
        for s in self.slices:
            if s.seq == slice_seq:
                s.fill_qty = fill_qty
                s.fill_price = fill_price
                s.status = "filled"
                break
        self.filled_qty += fill_qty
        self.total_cost += fill_qty * fill_price
        self.avg_fill_price = self.total_cost / self.filled_qty if self.filled_qty > 0 else 0
        if self.filled_qty >= self.total_qty:
            self.status = "completed"


class ExecAlgorithm(ABC):
    """Base execution algorithm.

    Subclass and implement ``generate_slices`` to define how a parent
    order is split into child orders.
    """

    name: str = "base"

    @abstractmethod
    def generate_slices(
        self, req: OrderSubmitRequest, params: Dict[str, Any]
    ) -> List[AlgoSlice]:
        """Split the parent order into child slices."""
        ...

    def on_fill(self, state: AlgoState, slice_seq: int, fill_qty: float, fill_price: float) -> None:
        """Called when a child slice is filled."""
        state.record_fill(slice_seq, fill_qty, fill_price)

    def on_cancel(self, state: AlgoState) -> None:
        """Cancel all remaining unfilled slices."""
        for s in state.slices:
            if s.status == "pending":
                s.status = "canceled"
        state.status = "canceled"



"""Tests for cpz.strategy — Strategy base class and lifecycle."""

from __future__ import annotations

from unittest.mock import MagicMock, patch

import pytest

from cpz.events.bus import EventBus
from cpz.model.events import BarEvent, FillEvent
from cpz.strategy.base import Strategy
from cpz.strategy.config import StrategyConfig
from cpz.strategy.context import Cache, Clock, Portfolio


class SimpleStrategy(Strategy):
    """Test strategy that records lifecycle calls."""

    def __init__(self, config=None):
        super().__init__(config)
        self.started = False
        self.stopped = False
        self.bars_received = []
        self.fills_received = []

    def on_start(self):
        self.started = True

    def on_stop(self):
        self.stopped = True

    def on_bar(self, bar):
        self.bars_received.append(bar)

    def on_order_filled(self, event):
        self.fills_received.append(event)


class TestStrategyConfig:
    def test_defaults(self):
        cfg = StrategyConfig()
        assert cfg.instruments == []
        assert cfg.broker == "alpaca"
        assert cfg.environment == "paper"

    def test_custom_fields(self):
        class MyConfig(StrategyConfig):
            fast: int = 10
            slow: int = 30

        cfg = MyConfig(instruments=["AAPL"], fast=5)
        assert cfg.fast == 5
        assert cfg.slow == 30
        assert cfg.instruments == ["AAPL"]

    def test_serialization(self):
        cfg = StrategyConfig(instruments=["AAPL"], strategy_id="test-123")
        d = cfg.to_dict()
        assert d["strategy_id"] == "test-123"
        assert d["instruments"] == ["AAPL"]

        restored = StrategyConfig.from_dict(d)
        assert restored.strategy_id == "test-123"


class TestStrategy:
    def test_auto_id(self):
        s = SimpleStrategy()
        assert s.strategy_id.startswith("SimpleStrategy-")

    def test_explicit_id(self):
        cfg = StrategyConfig(strategy_id="my-strat-001")
        s = SimpleStrategy(cfg)
        assert s.strategy_id == "my-strat-001"

    def test_lifecycle(self):
        s = SimpleStrategy()
        s.on_start()
        assert s.started
        s.on_stop()
        assert s.stopped

    def test_bar_handler(self):
        s = SimpleStrategy()
        bar = BarEvent(symbol="AAPL", close=150.0)
        s._handle_bar(bar)
        assert len(s.bars_received) == 1
        assert s.bars_received[0].symbol == "AAPL"

    def test_bar_cached(self):
        s = SimpleStrategy()
        bar = BarEvent(symbol="AAPL", close=150.0)
        s._handle_bar(bar)
        cached = s.cache.bars("AAPL")
        assert len(cached) == 1
        assert cached[0].close == 150.0

    def test_fill_updates_portfolio(self):
        s = SimpleStrategy()
        fill = FillEvent(
            symbol="AAPL", side="buy", fill_qty=10, fill_price=150.0,
            commission=0.5, strategy_id=s.strategy_id,
        )
        s._handle_fill(fill)
        assert not s.portfolio.is_flat("AAPL")
        assert s.portfolio.is_long("AAPL")
        assert len(s.fills_received) == 1

    def test_subscribe_bars(self):
        s = SimpleStrategy()
        s.subscribe_bars("AAPL", "1D")
        assert len(s._subscriptions) == 1
        assert s._subscriptions[0] == {"type": "bar", "symbol": "AAPL", "timeframe": "1D"}

    def test_subscribe_bars_dedup(self):
        s = SimpleStrategy()
        s.subscribe_bars("AAPL", "1D")
        s.subscribe_bars("AAPL", "1D")
        assert len(s._subscriptions) == 1

    def test_buy_sell_convenience(self):
        s = SimpleStrategy()
        bus = EventBus()
        s._bus = bus
        submitted = []
        bus.subscribe("order.submit", lambda req: submitted.append(req))

        s.buy("AAPL", qty=10)
        assert len(submitted) == 1
        assert submitted[0].side.value == "buy"

        s.sell("AAPL", qty=10)
        assert len(submitted) == 2
        assert submitted[1].side.value == "sell"

    def test_flatten(self):
        s = SimpleStrategy()
        bus = EventBus()
        s._bus = bus

        fill = FillEvent(
            symbol="AAPL", side="buy", fill_qty=10, fill_price=150.0,
            commission=0, strategy_id=s.strategy_id,
        )
        s._handle_fill(fill)
        assert s.portfolio.is_long("AAPL")

        submitted = []
        bus.subscribe("order.submit", lambda req: submitted.append(req))
        s.flatten("AAPL")
        assert len(submitted) == 1
        assert submitted[0].side.value == "sell"
        assert submitted[0].qty == 10.0

    def test_flatten_when_flat_noop(self):
        s = SimpleStrategy()
        result = s.flatten("AAPL")
        assert result is None

    def test_emit_signal(self):
        s = SimpleStrategy()
        bus = EventBus()
        s._bus = bus
        received = []
        bus.subscribe("signal.my_signal", lambda sig: received.append(sig))

        s.emit_signal("my_signal", value=1.5, data={"key": "val"})
        assert len(received) == 1
        assert received[0].name == "my_signal"
        assert received[0].value == 1.5


class TestPortfolio:
    def test_initial_state(self):
        p = Portfolio(100_000)
        assert p.cash == 100_000
        assert p.equity() == 100_000
        assert p.is_completely_flat()

    def test_apply_fill_buy(self):
        p = Portfolio(100_000)
        fill = FillEvent(symbol="AAPL", side="buy", fill_qty=10, fill_price=150.0, commission=1.0)
        p.apply_fill(fill)
        assert p.is_long("AAPL")
        assert p.cash == 100_000 - 10 * 150 - 1.0
        assert p.commission_total == 1.0

    def test_apply_fill_sell_closes(self):
        p = Portfolio(100_000)
        buy = FillEvent(symbol="AAPL", side="buy", fill_qty=10, fill_price=150.0, commission=0)
        p.apply_fill(buy)

        sell = FillEvent(symbol="AAPL", side="sell", fill_qty=10, fill_price=160.0, commission=0)
        p.apply_fill(sell)
        assert p.is_flat("AAPL")
        assert p.realized_pnl == 100.0  # 10 * (160 - 150)

    def test_positions(self):
        p = Portfolio(100_000)
        fill = FillEvent(symbol="AAPL", side="buy", fill_qty=5, fill_price=100.0, commission=0)
        p.apply_fill(fill)
        positions = p.positions()
        assert "AAPL" in positions
        assert positions["AAPL"].qty == 5

    def test_reset(self):
        p = Portfolio(100_000)
        fill = FillEvent(symbol="AAPL", side="buy", fill_qty=5, fill_price=100.0, commission=0)
        p.apply_fill(fill)
        p.reset()
        assert p.cash == 100_000
        assert p.is_completely_flat()


class TestClock:
    def test_real_time(self):
        c = Clock()
        assert not c.is_simulated
        ts = c.timestamp_ns()
        assert ts > 0

    def test_simulated(self):
        c = Clock()
        c.set_time(1_000_000_000_000)
        assert c.is_simulated
        assert c.timestamp_ns() == 1_000_000_000_000

    def test_reset(self):
        c = Clock()
        c.set_time(1_000)
        c.reset()
        assert not c.is_simulated


class TestCache:
    def test_push_and_retrieve_bars(self):
        c = Cache()
        bar1 = BarEvent(symbol="AAPL", close=150.0)
        bar2 = BarEvent(symbol="AAPL", close=151.0)
        c.push_bar(bar1)
        c.push_bar(bar2)
        bars = c.bars("AAPL")
        assert len(bars) == 2
        assert bars[-1].close == 151.0

    def test_latest_bar(self):
        c = Cache()
        c.push_bar(BarEvent(symbol="AAPL", close=150.0))
        c.push_bar(BarEvent(symbol="AAPL", close=155.0))
        assert c.latest_bar("AAPL").close == 155.0

    def test_empty_latest_bar(self):
        c = Cache()
        assert c.latest_bar("AAPL") is None

    def test_custom_storage(self):
        c = Cache()
        c.set("key1", "value1")
        assert c.get("key1") == "value1"
        assert c.get("missing", "default") == "default"



"""Rust acceleration for indicator computation.

Mirrors the pattern in ``cpz.risk._rust_accel``: try to import the
compiled extension, expose thin wrappers that return ``None`` when Rust
is unavailable so callers can fall back to Python/NumPy.

Build with:  ``cd rust && maturin develop --release``
"""

from __future__ import annotations

from typing import List, Optional, Tuple

try:
    import cpz_risk_rs

    RUST_AVAILABLE = True
except ImportError:
    RUST_AVAILABLE = False


def has_rust() -> bool:
    """``True`` when the compiled Rust extension is installed."""
    return RUST_AVAILABLE


# ── Trend ────────────────────────────────────────────────────────────

def fast_sma(close: List[float], period: int) -> Optional[List[float]]:
    if not RUST_AVAILABLE or not hasattr(cpz_risk_rs, "rolling_sma"):
        return None
    return cpz_risk_rs.rolling_sma(close, period)


def fast_ema(close: List[float], period: int) -> Optional[List[float]]:
    if not RUST_AVAILABLE or not hasattr(cpz_risk_rs, "ema"):
        return None
    return cpz_risk_rs.ema(close, period)


def fast_wma(close: List[float], period: int) -> Optional[List[float]]:
    if not RUST_AVAILABLE or not hasattr(cpz_risk_rs, "wma"):
        return None
    return cpz_risk_rs.wma(close, period)


def fast_kama(
    close: List[float], period: int, fast_sc: float, slow_sc: float,
) -> Optional[List[float]]:
    if not RUST_AVAILABLE or not hasattr(cpz_risk_rs, "kama"):
        return None
    return cpz_risk_rs.kama(close, period, fast_sc, slow_sc)


# ── Momentum ─────────────────────────────────────────────────────────

def fast_rsi(close: List[float], period: int) -> Optional[List[float]]:
    if not RUST_AVAILABLE or not hasattr(cpz_risk_rs, "rsi"):
        return None
    return cpz_risk_rs.rsi(close, period)


def fast_macd(
    close: List[float], fast: int, slow: int, signal: int,
) -> Optional[Tuple[List[float], List[float], List[float]]]:
    if not RUST_AVAILABLE or not hasattr(cpz_risk_rs, "macd"):
        return None
    return cpz_risk_rs.macd(close, fast, slow, signal)


def fast_adx(
    high: List[float], low: List[float], close: List[float], period: int,
) -> Optional[List[float]]:
    if not RUST_AVAILABLE or not hasattr(cpz_risk_rs, "adx"):
        return None
    return cpz_risk_rs.adx(high, low, close, period)


# ── Volatility ───────────────────────────────────────────────────────

def fast_atr(
    high: List[float], low: List[float], close: List[float], period: int,
) -> Optional[List[float]]:
    if not RUST_AVAILABLE or not hasattr(cpz_risk_rs, "atr"):
        return None
    return cpz_risk_rs.atr(high, low, close, period)


def fast_bollinger(
    close: List[float], period: int, num_std: float,
) -> Optional[Tuple[List[float], List[float], List[float]]]:
    if not RUST_AVAILABLE or not hasattr(cpz_risk_rs, "bollinger"):
        return None
    return cpz_risk_rs.bollinger(close, period, num_std)


def fast_garman_klass(
    open_: List[float],
    high: List[float],
    low: List[float],
    close: List[float],
    period: int,
) -> Optional[List[float]]:
    if not RUST_AVAILABLE or not hasattr(cpz_risk_rs, "garman_klass_vol"):
        return None
    return cpz_risk_rs.garman_klass_vol(open_, high, low, close, period)



"""StrategyContext — provides Clock, Cache, and Portfolio abstractions.

In live mode these delegate to the real broker. In backtest mode they
operate on simulated state. Strategies never touch the context internals
directly — they use the public API on the Strategy base class.
"""

from __future__ import annotations

import time
from collections import defaultdict, deque
from dataclasses import dataclass, field
from datetime import datetime, timezone
from typing import Any, Deque, Dict, List, Optional

from ..model.events import BarEvent, FillEvent


class Clock:
    """System clock — real-time in live mode, simulated in backtest."""

    def __init__(self) -> None:
        self._simulated_ns: Optional[int] = None

    def set_time(self, ts_ns: int) -> None:
        self._simulated_ns = ts_ns

    def reset(self) -> None:
        self._simulated_ns = None

    def timestamp_ns(self) -> int:
        if self._simulated_ns is not None:
            return self._simulated_ns
        return int(time.time() * 1_000_000_000)

    def utc_now(self) -> datetime:
        ns = self.timestamp_ns()
        return datetime.fromtimestamp(ns / 1e9, tz=timezone.utc)

    @property
    def is_simulated(self) -> bool:
        return self._simulated_ns is not None


class Cache:
    """In-memory cache for bars, quotes, orders, and positions."""

    def __init__(self, max_bars: int = 500) -> None:
        self._max_bars = max_bars
        self._bars: Dict[str, Deque[BarEvent]] = defaultdict(lambda: deque(maxlen=max_bars))
        self._last_quote: Dict[str, Any] = {}
        self._orders: Dict[str, Any] = {}
        self._custom: Dict[str, Any] = {}

    def push_bar(self, bar: BarEvent) -> None:
        self._bars[bar.symbol].append(bar)

    def bars(self, symbol: str, limit: int | None = None) -> List[BarEvent]:
        all_bars = list(self._bars.get(symbol, []))
        if limit:
            return all_bars[-limit:]
        return all_bars

    def latest_bar(self, symbol: str) -> Optional[BarEvent]:
        dq = self._bars.get(symbol)
        return dq[-1] if dq else None

    def set_quote(self, symbol: str, quote: Any) -> None:
        self._last_quote[symbol] = quote

    def quote(self, symbol: str) -> Any:
        return self._last_quote.get(symbol)

    def set_order(self, order_id: str, order: Any) -> None:
        self._orders[order_id] = order

    def order(self, order_id: str) -> Any:
        return self._orders.get(order_id)

    def set(self, key: str, value: Any) -> None:
        self._custom[key] = value

    def get(self, key: str, default: Any = None) -> Any:
        return self._custom.get(key, default)

    def clear(self) -> None:
        self._bars.clear()
        self._last_quote.clear()
        self._orders.clear()
        self._custom.clear()


@dataclass
class PositionState:
    """Tracked position for a single symbol."""

    symbol: str = ""
    qty: float = 0.0
    avg_entry_price: float = 0.0
    realized_pnl: float = 0.0
    cost_basis: float = 0.0

    @property
    def is_long(self) -> bool:
        return self.qty > 0

    @property
    def is_short(self) -> bool:
        return self.qty < 0

    @property
    def is_flat(self) -> bool:
        return self.qty == 0

    def market_value(self, current_price: float) -> float:
        return self.qty * current_price

    def unrealized_pnl(self, current_price: float) -> float:
        if self.qty == 0:
            return 0.0
        return (current_price - self.avg_entry_price) * self.qty


class Portfolio:
    """Portfolio state — positions, equity, P&L tracking."""

    def __init__(self, initial_cash: float = 100_000.0) -> None:
        self._initial_cash = initial_cash
        self._cash = initial_cash
        self._positions: Dict[str, PositionState] = {}
        self._realized_pnl = 0.0
        self._commission_total = 0.0
        self._last_prices: Dict[str, float] = {}

    @property
    def cash(self) -> float:
        return self._cash

    @property
    def initial_cash(self) -> float:
        return self._initial_cash

    @property
    def realized_pnl(self) -> float:
        return self._realized_pnl

    @property
    def commission_total(self) -> float:
        return self._commission_total

    def equity(self) -> float:
        unrealized = sum(
            pos.unrealized_pnl(self._last_prices.get(sym, pos.avg_entry_price))
            for sym, pos in self._positions.items()
            if not pos.is_flat
        )
        return self._cash + sum(
            pos.market_value(self._last_prices.get(sym, pos.avg_entry_price))
            for sym, pos in self._positions.items()
            if not pos.is_flat
        )

    def update_price(self, symbol: str, price: float) -> None:
        self._last_prices[symbol] = price

    def position(self, symbol: str) -> Optional[PositionState]:
        pos = self._positions.get(symbol)
        if pos and pos.is_flat:
            return None
        return pos

    def positions(self) -> Dict[str, PositionState]:
        return {s: p for s, p in self._positions.items() if not p.is_flat}

    def is_long(self, symbol: str) -> bool:
        pos = self._positions.get(symbol)
        return pos is not None and pos.is_long

    def is_short(self, symbol: str) -> bool:
        pos = self._positions.get(symbol)
        return pos is not None and pos.is_short

    def is_flat(self, symbol: str) -> bool:
        pos = self._positions.get(symbol)
        return pos is None or pos.is_flat

    def is_completely_flat(self) -> bool:
        return all(p.is_flat for p in self._positions.values())

    def apply_fill(self, fill: FillEvent) -> None:
        """Update position state from a fill event."""
        sym = fill.symbol
        pos = self._positions.get(sym)
        if pos is None:
            pos = PositionState(symbol=sym)
            self._positions[sym] = pos

        signed_qty = fill.fill_qty if fill.side == "buy" else -fill.fill_qty
        cost = fill.fill_qty * fill.fill_price
        commission = fill.commission

        if pos.is_flat:
            pos.qty = signed_qty
            pos.avg_entry_price = fill.fill_price
            pos.cost_basis = cost
        elif (pos.qty > 0 and signed_qty > 0) or (pos.qty < 0 and signed_qty < 0):
            total_cost = pos.cost_basis + cost
            pos.qty += signed_qty
            pos.cost_basis = total_cost
            pos.avg_entry_price = total_cost / abs(pos.qty) if pos.qty != 0 else 0
        else:
            close_qty = min(abs(signed_qty), abs(pos.qty))
            pnl = close_qty * (fill.fill_price - pos.avg_entry_price) * (1 if pos.qty > 0 else -1)
            pos.realized_pnl += pnl
            self._realized_pnl += pnl
            pos.qty += signed_qty
            if abs(pos.qty) < 1e-10:
                pos.qty = 0.0
                pos.avg_entry_price = 0.0
                pos.cost_basis = 0.0
            else:
                pos.cost_basis = abs(pos.qty) * pos.avg_entry_price

        self._cash -= (cost if signed_qty > 0 else -cost)
        self._cash -= commission
        self._commission_total += commission
        self._last_prices[sym] = fill.fill_price

    def reset(self, initial_cash: float | None = None) -> None:
        if initial_cash is not None:
            self._initial_cash = initial_cash
        self._cash = self._initial_cash
        self._positions.clear()
        self._realized_pnl = 0.0
        self._commission_total = 0.0
        self._last_prices.clear()



"""Finnhub data provider.

Real-time stock data, news, earnings, and analyst recommendations.
Free tier: 60 API calls/minute.

Docs: https://finnhub.io/docs/api
"""

from __future__ import annotations

from datetime import datetime, timedelta
from typing import Any, Dict, List, Optional

import requests

from ..models import Bar, Quote, TimeFrame


class FinnhubProvider:
    """Finnhub data provider.
    
    Supports stocks, news, earnings, and analyst recommendations.
    
    Usage:
        provider = FinnhubProvider(api_key="your_key")
        bars = provider.get_bars("AAPL", start=datetime(2023, 1, 1))
    """
    
    name = "finnhub"
    BASE_URL = "https://finnhub.io"
    
    RESOLUTION_MAP = {
        TimeFrame.MINUTE_1: "1",
        TimeFrame.MINUTE_5: "5",
        TimeFrame.MINUTE_15: "15",
        TimeFrame.MINUTE_30: "30",
        TimeFrame.HOUR_1: "60",
        TimeFrame.HOUR_4: "D",
        TimeFrame.DAY: "D",
        TimeFrame.WEEK: "W",
        TimeFrame.MONTH: "M",
    }
    
    def __init__(self, api_key: Optional[str] = None) -> None:
        """Initialize Finnhub provider.
        
        Args:
            api_key: Finnhub API key. Get free key at finnhub.io
        """
        self._api_key = api_key or ""
        
        if not self._api_key:
            try:
                from ...common.cpz_ai import CPZAIClient
                cpz_client = CPZAIClient.from_env()
                creds = cpz_client.get_data_credentials(provider="finnhub")
                self._api_key = creds.get("finnhub_api_key", "")
            except Exception:
                pass
        
        if not self._api_key:
            print("[CPZ SDK] Finnhub API key not found. Get a free key at finnhub.io")
    
    def _request(self, path: str, params: Optional[Dict[str, Any]] = None) -> Any:
        """Make API request with token auth."""
        if params is None:
            params = {}
        params["token"] = self._api_key
        
        url = f"{self.BASE_URL}{path}"
        response = requests.get(url, params=params, timeout=30)
        response.raise_for_status()
        return response.json()
    
    def get_bars(
        self,
        symbol: str,
        timeframe: TimeFrame | str = TimeFrame.DAY,
        start: Optional[datetime] = None,
        end: Optional[datetime] = None,
        limit: Optional[int] = None,
    ) -> List[Bar]:
        """Get historical candle data from Finnhub.
        
        Args:
            symbol: Stock symbol (e.g., "AAPL")
            timeframe: Bar timeframe
            start: Start datetime
            end: End datetime
            limit: Maximum bars
            
        Returns:
            List of Bar objects
        """
        if isinstance(timeframe, str):
            timeframe = TimeFrame.from_string(timeframe)
        
        resolution = self.RESOLUTION_MAP.get(timeframe, "D")
        
        if not start:
            start = datetime.utcnow() - timedelta(days=365)
        if not end:
            end = datetime.utcnow()
        
        from_ts = int(start.timestamp())
        to_ts = int(end.timestamp())
        
        try:
            data = self._request("/api/v1/stock/candle", params={
                "symbol": symbol,
                "resolution": resolution,
                "from": from_ts,
                "to": to_ts,
            })
            
            if data.get("s") != "ok":
                print(f"[CPZ SDK] No Finnhub data for {symbol}: {data.get('s', 'unknown')}")
                return []
            
            timestamps = data.get("t", [])
            opens = data.get("o", [])
            highs = data.get("h", [])
            lows = data.get("l", [])
            closes = data.get("c", [])
            volumes = data.get("v", [])
            
            bars: List[Bar] = []
            for i in range(len(timestamps)):
                bars.append(Bar(
                    symbol=symbol,
                    timestamp=datetime.utcfromtimestamp(timestamps[i]),
                    open=float(opens[i]),
                    high=float(highs[i]),
                    low=float(lows[i]),
                    close=float(closes[i]),
                    volume=float(volumes[i]) if i < len(volumes) else 0.0,
                    vwap=None,
                ))
            
            bars.sort(key=lambda x: x.timestamp)
            
            if limit and len(bars) > limit:
                bars = bars[-limit:]
            
            return bars
            
        except Exception as e:
            print(f"[CPZ SDK] Error fetching Finnhub data for {symbol}: {e}")
            return []
    
    def get_quote(self, symbol: str) -> Optional[Quote]:
        """Get latest quote from Finnhub.
        
        Response fields: c (current), h (high), l (low), o (open), pc (prev close), t (timestamp).
        """
        try:
            data = self._request("/api/v1/quote", params={"symbol": symbol})
            
            price = data.get("c")
            if price is None or price == 0:
                return None
            
            return Quote(
                symbol=symbol,
                timestamp=datetime.utcfromtimestamp(data.get("t", 0)) if data.get("t") else datetime.utcnow(),
                bid=float(price),
                ask=float(price),
                bid_size=0,
                ask_size=0,
            )
            
        except Exception as e:
            print(f"[CPZ SDK] Error fetching Finnhub quote for {symbol}: {e}")
            return None
    
    def get_quotes(self, symbols: List[str]) -> List[Quote]:
        """Get quotes for multiple symbols."""
        quotes: List[Quote] = []
        for symbol in symbols:
            quote = self.get_quote(symbol)
            if quote:
                quotes.append(quote)
        return quotes
    
    def get_news(
        self,
        symbol: str,
        start: Optional[datetime] = None,
        end: Optional[datetime] = None,
    ) -> List[Dict[str, Any]]:
        """Get company news from Finnhub.
        
        Args:
            symbol: Stock symbol
            start: Start date
            end: End date
            
        Returns:
            List of news article dicts
        """
        if not start:
            start = datetime.utcnow() - timedelta(days=7)
        if not end:
            end = datetime.utcnow()
        
        try:
            return self._request("/api/v1/company-news", params={
                "symbol": symbol,
                "from": start.strftime("%Y-%m-%d"),
                "to": end.strftime("%Y-%m-%d"),
            })
        except Exception as e:
            print(f"[CPZ SDK] Error fetching Finnhub news for {symbol}: {e}")
            return []
    
    def get_recommendation(self, symbol: str) -> List[Dict[str, Any]]:
        """Get analyst recommendation trends.
        
        Args:
            symbol: Stock symbol
            
        Returns:
            List of recommendation dicts with buy/hold/sell/strongBuy/strongSell counts
        """
        try:
            return self._request("/api/v1/stock/recommendation", params={"symbol": symbol})
        except Exception as e:
            print(f"[CPZ SDK] Error fetching Finnhub recommendations for {symbol}: {e}")
            return []
    
    def get_earnings(self, symbol: str) -> List[Dict[str, Any]]:
        """Get company earnings surprises.
        
        Args:
            symbol: Stock symbol
            
        Returns:
            List of earnings dicts with actual/estimate/period/surprise
        """
        try:
            return self._request("/api/v1/stock/earnings", params={"symbol": symbol})
        except Exception as e:
            print(f"[CPZ SDK] Error fetching Finnhub earnings for {symbol}: {e}")
            return []



"""
Polymarket Adapter
==================
BrokerAdapter implementation for Polymarket prediction markets.

All operations are proxied through the polymarket-direct Supabase Edge Function
which handles wallet credential decryption and CLOB API auth server-side.
"""

from __future__ import annotations

import os
import logging
from datetime import datetime
from typing import AsyncIterator, Dict, Any, Iterable, List, Optional

import requests

from ..enums import OrderSide, OrderType, TimeInForce
from ..interfaces import BrokerAdapter
from ..models import (
    OrderSubmitRequest,
    OrderReplaceRequest,
    Order,
    Account,
    Position,
    Quote,
    Bar,
)

logger = logging.getLogger(__name__)

SUPABASE_URL = "https://brkcjojfmlygsujiqglv.supabase.co"
SUPABASE_ANON_KEY = (
    "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9."
    "eyJpc3MiOiJzdXBhYmFzZSIsInJlZiI6ImJya2Nqb2pmbWx5Z3N1amlxZ2x2Iiwicm9sZSI6ImFub24iLCJpYXQiOjE3NDc5MDg3NTksImV4cCI6MjA2MzQ4NDc1OX0."
    "1mM0avWJHVH-7_MKIW7_gvh-WwMj97qTVOe0pbk0S-E"
)


class PolymarketAdapter(BrokerAdapter):
    """
    Polymarket adapter for CPZ Python SDK.

    Connects to the Polymarket prediction market through the
    ``polymarket-direct`` Supabase Edge Function. The edge function
    handles wallet credential decryption and CLOB API authentication.

    Prerequisites:
        1. Connect your Polygon wallet via the CPZAI web UI
           (Settings -> Connect Broker -> Polymarket).
        2. The connected wallet will appear in your trading credentials.

    Usage::

        from cpz import CPZClient

        client = CPZClient()
        client.execution.use_broker("polymarket")

        account = client.execution.get_account()
        positions = client.execution.get_positions()

        order = client.execution.submit_order(
            OrderSubmitRequest(
                symbol="market-condition-id",
                side=OrderSide.BUY,
                qty=100,
                type=OrderType.LIMIT,
                limit_price=0.65,
                outcome="YES",
            )
        )
    """

    def __init__(
        self,
        cpz_api_key: str,
        account_id: str,
        supabase_url: str = SUPABASE_URL,
        supabase_anon_key: str = SUPABASE_ANON_KEY,
        timeout: int = 30,
    ) -> None:
        self._cpz_api_key = cpz_api_key
        self._account_id = account_id
        self._supabase_url = supabase_url.rstrip("/")
        self._anon_key = supabase_anon_key
        self._timeout = timeout
        self._fn_url = f"{self._supabase_url}/functions/v1/polymarket-direct"

    @staticmethod
    def create(**kwargs: object) -> "PolymarketAdapter":
        from ...common.cpz_ai import CPZAIClient

        account_id = str(kwargs.get("account_id") or "")
        env = str(kwargs.get("env") or os.getenv("POLYMARKET_ENV", "live"))
        supabase_url = str(kwargs.get("supabase_url") or os.getenv("SUPABASE_URL", SUPABASE_URL))
        supabase_anon_key = str(
            kwargs.get("supabase_anon_key")
            or os.getenv("SUPABASE_ANON_KEY", SUPABASE_ANON_KEY)
        )

        platform = CPZAIClient.from_env()
        cpz_api_key = platform.api_key

        if not account_id:
            try:
                creds = platform.get_broker_credentials(
                    broker="polymarket", env=env or None, account_id=None
                )
                if creds:
                    account_id = str(creds.get("id") or creds.get("account_id") or "")
            except Exception as exc:
                logger.warning("Failed to get Polymarket credentials from platform: %s", exc)

        if not account_id:
            raise ValueError(
                "Polymarket wallet not found. Connect your Polygon wallet via the CPZAI "
                "web UI first (Settings -> Connect Broker -> Polymarket)."
            )

        return PolymarketAdapter(
            cpz_api_key=cpz_api_key,
            account_id=account_id,
            supabase_url=supabase_url,
            supabase_anon_key=supabase_anon_key,
        )

    def _call(self, action: str, extra: Optional[Dict[str, Any]] = None) -> Any:
        """Call the polymarket-direct edge function."""
        payload: Dict[str, Any] = {"action": action, "accountId": self._account_id}
        if extra:
            payload.update(extra)

        headers = {
            "apikey": self._anon_key,
            "X-CPZ-Key": self._cpz_api_key,
            "Content-Type": "application/json",
        }

        try:
            resp = requests.post(
                self._fn_url, json=payload, headers=headers, timeout=self._timeout
            )
        except requests.exceptions.Timeout:
            raise Exception(f"Polymarket edge function timeout after {self._timeout}s")
        except requests.exceptions.ConnectionError:
            raise Exception(f"Failed to connect to Polymarket edge function at {self._fn_url}")

        if resp.status_code >= 400:
            error_text = resp.text
            try:
                error_data = resp.json()
                error_text = error_data.get("message") or error_data.get("error") or error_text
            except Exception:
                pass
            raise Exception(f"Polymarket error ({resp.status_code}): {error_text}")

        data = resp.json()
        if data.get("error"):
            raise Exception(data.get("message") or data["error"])
        return data.get("data", data)

    # BrokerAdapter implementation

    def get_account(self) -> Account:
        data = self._call("account")
        return Account(
            id=str(data.get("wallet_address") or data.get("account_id") or self._account_id),
            buying_power=float(data.get("buying_power") or 0),
            equity=float(data.get("equity") or data.get("portfolio_value") or 0),
            cash=float(data.get("cash") or data.get("buying_power") or 0),
        )

    def get_positions(self) -> list[Position]:
        data = self._call("positions")
        if not isinstance(data, list):
            data = []
        return [
            Position(
                symbol=str(p.get("symbol", "")),
                qty=float(p.get("qty") or 0),
                avg_entry_price=float(p.get("avg_entry_price") or 0),
                market_value=float(p.get("market_value") or 0),
            )
            for p in data
        ]

    def get_orders(self, status: Optional[str] = None, limit: int = 100) -> list[Order]:
        data = self._call("orders")
        if not isinstance(data, list):
            data = []

        side_map = {"buy": OrderSide.BUY, "sell": OrderSide.SELL}
        type_map = {"market": OrderType.MARKET, "limit": OrderType.LIMIT}

        orders = []
        for o in data:
            order_status = str(o.get("status", "unknown"))
            if status and order_status != status:
                continue
            raw_side = str(o.get("side", "buy")).lower()
            raw_type = str(o.get("type", "limit")).lower()
            orders.append(
                Order(
                    id=str(o.get("id", "")),
                    symbol=str(o.get("symbol", "")),
                    side=side_map.get(raw_side, OrderSide.BUY),
                    qty=float(o.get("qty") or 0),
                    filled_qty=float(o.get("filled_qty") or 0),
                    type=type_map.get(raw_type, OrderType.LIMIT),
                    time_in_force=TimeInForce.GTC,
                    status=order_status,
                    created_at=datetime.utcnow(),
                )
            )
        return orders[:limit]

    def submit_order(self, request: OrderSubmitRequest) -> Order:
        order_payload: Dict[str, Any] = {
            "token_id": getattr(request, "token_id", None) or request.symbol,
            "polymarket_token_id": getattr(request, "token_id", None) or request.symbol,
            "side": request.side.value if hasattr(request.side, "value") else str(request.side),
            "qty": request.qty,
            "type": request.type.value if hasattr(request.type, "value") else str(request.type),
            "outcome": getattr(request, "outcome", "YES"),
        }
        if request.limit_price is not None:
            order_payload["limit_price"] = request.limit_price

        result = self._call("place_order", {"order": order_payload})
        return Order(
            id=str(result.get("id") or result.get("orderID") or ""),
            symbol=request.symbol,
            side=request.side,
            qty=request.qty or 0,
            filled_qty=0,
            type=request.order_type,
            time_in_force=request.time_in_force,
            status=str(result.get("status", "new")),
            created_at=datetime.utcnow(),
        )

    def cancel_order(self, order_id: str) -> None:
        self._call("cancel_order", {"orderId": order_id})

    def replace_order(self, order_id: str, request: OrderReplaceRequest) -> Order:
        self.cancel_order(order_id)
        submit_req = OrderSubmitRequest(
            symbol=request.symbol or "",
            side=OrderSide.BUY,
            qty=request.qty or 0,
            type=OrderType.LIMIT,
            limit_price=request.limit_price,
        )
        return self.submit_order(submit_req)

    def get_order(self, order_id: str) -> Order:
        orders = self.get_orders()
        for o in orders:
            if o.id == order_id:
                return o
        raise ValueError(f"Order {order_id} not found")

    def cancel_all_orders(self) -> int:
        orders = self.get_orders(status="new")
        for o in orders:
            try:
                self.cancel_order(o.id)
            except Exception:
                pass
        return len(orders)

    def get_quote(self, symbol: str) -> Quote:
        raise NotImplementedError("Use client.data.polymarket for market quotes")

    def get_bars(self, symbol: str, timeframe: str = "1Day", limit: int = 100) -> list[Bar]:
        raise NotImplementedError("Use client.data.polymarket for market data")

    async def stream_quotes(self, symbols: Iterable[str]) -> AsyncIterator[Quote]:
        raise NotImplementedError("Polymarket streaming not yet supported")

    async def stream_bars(self, symbols: Iterable[str], timeframe: str = "1Min") -> AsyncIterator[Bar]:
        raise NotImplementedError("Polymarket streaming not yet supported")



"""Transaction cost models: Almgren-Chriss, linear/sqrt impact, spread, turnover analysis.

All functions are pure: data in, results out.
"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional

import numpy as np

EPSILON: float = 1e-15


@dataclass
class CostEstimate:
    """Breakdown of transaction cost for a single trade."""
    permanent_cost_bps: float = 0.0
    temporary_cost_bps: float = 0.0
    spread_cost_bps: float = 0.0
    commission_bps: float = 0.0
    total_cost_bps: float = 0.0


@dataclass
class TurnoverResult:
    """Portfolio turnover analysis."""
    one_way_turnover: float = 0.0
    two_way_turnover: float = 0.0
    estimated_cost_bps: float = 0.0
    per_asset_turnover: Dict[str, float] = field(default_factory=dict)
    per_asset_cost_bps: Dict[str, float] = field(default_factory=dict)


@dataclass
class CostModel:
    """Pluggable cost model for portfolio optimisation.

    All parameters are configurable so users can calibrate to their
    broker and market conditions.
    """
    spread_bps: float = 5.0
    impact_coefficient: float = 0.1
    impact_exponent: float = 0.5
    impact_model: str = "sqrt"
    commission_bps: float = 0.0
    permanent_fraction: float = 0.5

    def estimate(
        self,
        notional: float,
        adv: float = 0.0,
        volatility: float = 0.0,
    ) -> float:
        """Estimate round-trip cost in basis points.

        Args:
            notional: Trade notional value.
            adv: Average daily volume (notional).
            volatility: Daily volatility of the asset.
        """
        spread = self.spread_bps / 2.0

        if self.impact_model == "sqrt":
            impact = sqrt_impact(notional, adv, coefficient=self.impact_coefficient,
                                 volatility=volatility, exponent=self.impact_exponent)
        elif self.impact_model == "linear":
            impact = linear_impact(notional, adv, coefficient=self.impact_coefficient)
        elif self.impact_model == "almgren_chriss":
            result = almgren_chriss(notional, adv, volatility)
            impact = result.permanent_cost_bps + result.temporary_cost_bps
        else:
            impact = 0.0

        return spread + impact + self.commission_bps


def almgren_chriss(
    notional: float,
    adv: float,
    volatility: float,
    *,
    gamma: float = 0.314,
    eta: float = 0.142,
    alpha: float = 0.6,
    beta: float = 0.6,
    participation_rate: float = 0.1,
) -> CostEstimate:
    """Almgren-Chriss (2001) market impact model.

    permanent_impact = gamma * sigma * (Q/V)^alpha
    temporary_impact = eta * sigma * (Q/(V*T))^beta

    Args:
        notional: Trade notional.
        adv: Average daily volume (notional).
        volatility: Daily price volatility.
        gamma: Permanent impact coefficient.
        eta: Temporary impact coefficient.
        alpha: Permanent impact exponent.
        beta: Temporary impact exponent.
        participation_rate: Fraction of ADV consumed per time unit.
    """
    if adv < EPSILON or volatility < EPSILON:
        return CostEstimate()

    q_over_v = notional / adv
    perm = gamma * volatility * (q_over_v ** alpha) * 10000.0

    execution_time = q_over_v / max(participation_rate, EPSILON)
    temp = eta * volatility * ((q_over_v / max(execution_time, EPSILON)) ** beta) * 10000.0

    return CostEstimate(
        permanent_cost_bps=round(perm, 4),
        temporary_cost_bps=round(temp, 4),
        total_cost_bps=round(perm + temp, 4),
    )


def linear_impact(
    notional: float,
    adv: float,
    *,
    coefficient: float = 0.1,
) -> float:
    """Linear market impact in basis points.

    ``cost_bps = coefficient * (notional / ADV) * 10000``

    Args:
        notional: Trade notional.
        adv: Average daily volume.
        coefficient: Impact coefficient (calibrate to your market).
    """
    if adv < EPSILON:
        return 0.0
    return coefficient * (notional / adv) * 10000.0


def sqrt_impact(
    notional: float,
    adv: float,
    *,
    coefficient: float = 0.1,
    volatility: float = 0.0,
    exponent: float = 0.5,
) -> float:
    """Square-root market impact (Torre 1997, Barra).

    ``cost_bps = coefficient * sigma * (notional / ADV)^exponent * 10000``

    Args:
        notional: Trade notional.
        adv: Average daily volume.
        coefficient: Impact coefficient.
        volatility: Daily price volatility (if 0, uses coefficient alone).
        exponent: Impact exponent (default 0.5 = square root).
    """
    if adv < EPSILON:
        return 0.0
    sigma = volatility if volatility > EPSILON else 1.0
    return coefficient * sigma * ((notional / adv) ** exponent) * 10000.0


def spread_cost(bid_ask_spread_bps: float) -> float:
    """Half-spread cost: ``cost = spread / 2``. The minimum cost of any trade."""
    return bid_ask_spread_bps / 2.0


def total_cost(
    notional: float,
    adv: float,
    volatility: float,
    *,
    bid_ask_bps: float = 5.0,
    commission_bps: float = 0.0,
    impact_model: str = "sqrt",
    impact_coefficient: float = 0.1,
    impact_exponent: float = 0.5,
) -> CostEstimate:
    """Combined transaction cost: spread + impact + commission.

    Args:
        notional: Trade notional.
        adv: Average daily volume.
        volatility: Daily price volatility.
        bid_ask_bps: Bid-ask spread in bps.
        commission_bps: Broker commission in bps.
        impact_model: ``"sqrt"`` (default), ``"linear"``, or ``"almgren_chriss"``.
        impact_coefficient: Impact coefficient.
        impact_exponent: Impact exponent (for sqrt model).
    """
    spread = spread_cost(bid_ask_bps)

    if impact_model == "sqrt":
        impact = sqrt_impact(notional, adv, coefficient=impact_coefficient,
                             volatility=volatility, exponent=impact_exponent)
        return CostEstimate(
            spread_cost_bps=round(spread, 4),
            temporary_cost_bps=round(impact, 4),
            commission_bps=commission_bps,
            total_cost_bps=round(spread + impact + commission_bps, 4),
        )
    elif impact_model == "linear":
        impact = linear_impact(notional, adv, coefficient=impact_coefficient)
        return CostEstimate(
            spread_cost_bps=round(spread, 4),
            temporary_cost_bps=round(impact, 4),
            commission_bps=commission_bps,
            total_cost_bps=round(spread + impact + commission_bps, 4),
        )
    elif impact_model == "almgren_chriss":
        ac = almgren_chriss(notional, adv, volatility)
        ac.spread_cost_bps = round(spread, 4)
        ac.commission_bps = commission_bps
        ac.total_cost_bps = round(
            ac.permanent_cost_bps + ac.temporary_cost_bps + spread + commission_bps, 4
        )
        return ac
    else:
        raise ValueError(f"Unknown impact_model '{impact_model}'. Use: sqrt, linear, almgren_chriss")


def turnover_analysis(
    current_weights: Dict[str, float],
    target_weights: Dict[str, float],
    *,
    cost_model: Optional[CostModel] = None,
    adv: Optional[Dict[str, float]] = None,
    volatility: Optional[Dict[str, float]] = None,
    portfolio_value: float = 1_000_000.0,
) -> TurnoverResult:
    """Analyse portfolio turnover and estimated rebalancing cost.

    Args:
        current_weights: ``{symbol: current_weight}``.
        target_weights: ``{symbol: target_weight}``.
        cost_model: Cost model for per-asset cost estimation.
        adv: ``{symbol: average_daily_volume}`` (optional).
        volatility: ``{symbol: daily_volatility}`` (optional).
        portfolio_value: Total portfolio value for notional calculation.
    """
    all_symbols = set(current_weights.keys()) | set(target_weights.keys())
    per_asset_turn: Dict[str, float] = {}
    per_asset_cost: Dict[str, float] = {}

    for sym in all_symbols:
        curr = current_weights.get(sym, 0.0)
        tgt = target_weights.get(sym, 0.0)
        turn = abs(tgt - curr)
        per_asset_turn[sym] = round(turn, 6)

        if cost_model and turn > EPSILON:
            notional = turn * portfolio_value
            sym_adv = (adv or {}).get(sym, portfolio_value * 0.01)
            sym_vol = (volatility or {}).get(sym, 0.02)
            per_asset_cost[sym] = round(cost_model.estimate(notional, sym_adv, sym_vol), 4)
        else:
            per_asset_cost[sym] = 0.0

    one_way = sum(per_asset_turn.values()) / 2.0
    two_way = sum(per_asset_turn.values())
    total_cost = sum(per_asset_cost.values())

    return TurnoverResult(
        one_way_turnover=round(one_way, 6),
        two_way_turnover=round(two_way, 6),
        estimated_cost_bps=round(total_cost, 4),
        per_asset_turnover=per_asset_turn,
        per_asset_cost_bps=per_asset_cost,
    )



"""Tests for cpz.execution.algos — TWAP, VWAP, Iceberg algorithms."""

from __future__ import annotations

import pytest

from cpz.execution.algos.base import AlgoState
from cpz.execution.algos.iceberg import IcebergAlgorithm
from cpz.execution.algos.twap import TWAPAlgorithm
from cpz.execution.algos.vwap import VWAPAlgorithm
from cpz.execution.enums import OrderSide, OrderType, TimeInForce
from cpz.execution.models import OrderSubmitRequest


def _order(qty=1000, symbol="AAPL", limit_price=None):
    otype = OrderType.LIMIT if limit_price else OrderType.MARKET
    return OrderSubmitRequest(
        symbol=symbol,
        side=OrderSide.BUY,
        qty=qty,
        order_type=otype,
        time_in_force=TimeInForce.DAY,
        limit_price=limit_price,
        strategy_id="test",
    )


class TestTWAP:
    def test_equal_slices(self):
        algo = TWAPAlgorithm()
        slices = algo.generate_slices(_order(qty=100), {"num_slices": 10, "duration_minutes": 10})
        assert len(slices) == 10
        total = sum(s.qty for s in slices)
        assert total == 100

    def test_remainder_distribution(self):
        algo = TWAPAlgorithm()
        slices = algo.generate_slices(_order(qty=103), {"num_slices": 10, "duration_minutes": 5})
        total = sum(s.qty for s in slices)
        assert total == 103

    def test_scheduled_times_ascending(self):
        algo = TWAPAlgorithm()
        slices = algo.generate_slices(_order(qty=100), {"num_slices": 5, "duration_minutes": 10})
        for i in range(1, len(slices)):
            assert slices[i].scheduled_time > slices[i - 1].scheduled_time

    def test_zero_qty_raises(self):
        algo = TWAPAlgorithm()
        with pytest.raises(ValueError, match="qty > 0"):
            algo.generate_slices(_order(qty=0), {"num_slices": 5})

    def test_single_slice(self):
        algo = TWAPAlgorithm()
        slices = algo.generate_slices(_order(qty=50), {"num_slices": 1, "duration_minutes": 1})
        assert len(slices) == 1
        assert slices[0].qty == 50


class TestVWAP:
    def test_default_profile(self):
        algo = VWAPAlgorithm()
        slices = algo.generate_slices(_order(qty=1000), {"duration_minutes": 60})
        total = sum(s.qty for s in slices)
        assert total == 1000
        assert len(slices) > 0

    def test_custom_profile(self):
        algo = VWAPAlgorithm()
        profile = [0.5, 0.3, 0.2]
        slices = algo.generate_slices(
            _order(qty=100),
            {"num_slices": 3, "volume_profile": profile, "duration_minutes": 30},
        )
        assert len(slices) == 3
        total = sum(s.qty for s in slices)
        assert total == 100

    def test_total_qty_preserved(self):
        algo = VWAPAlgorithm()
        slices = algo.generate_slices(
            _order(qty=777),
            {"num_slices": 5, "duration_minutes": 60},
        )
        total = sum(s.qty for s in slices)
        assert total == 777


class TestIceberg:
    def test_basic_split(self):
        algo = IcebergAlgorithm()
        slices = algo.generate_slices(_order(qty=500), {"display_qty": 100})
        assert len(slices) == 5
        assert all(s.qty == 100 for s in slices)

    def test_first_pending_rest_held(self):
        algo = IcebergAlgorithm()
        slices = algo.generate_slices(_order(qty=300), {"display_qty": 100})
        assert slices[0].status == "pending"
        assert slices[1].status == "held"
        assert slices[2].status == "held"

    def test_on_fill_releases_next(self):
        algo = IcebergAlgorithm()
        slices = algo.generate_slices(_order(qty=300), {"display_qty": 100})
        state = AlgoState(total_qty=300, slices=slices)

        algo.on_fill(state, 0, 100, 150.0)
        assert slices[1].status == "pending"
        assert slices[2].status == "held"

    def test_remainder_slice(self):
        algo = IcebergAlgorithm()
        slices = algo.generate_slices(_order(qty=250), {"display_qty": 100})
        assert len(slices) == 3
        assert slices[2].qty == 50

    def test_zero_display_qty_raises(self):
        algo = IcebergAlgorithm()
        with pytest.raises(ValueError, match="display_qty"):
            algo.generate_slices(_order(qty=500), {"display_qty": 0})

    def test_limit_order_preserved(self):
        algo = IcebergAlgorithm()
        slices = algo.generate_slices(_order(qty=200, limit_price=150.0), {"display_qty": 100})
        assert all(s.limit_price == 150.0 for s in slices)
        assert all(s.order_type == "limit" for s in slices)


class TestAlgoState:
    def test_remaining_qty(self):
        state = AlgoState(total_qty=100, filled_qty=30)
        assert state.remaining_qty == 70

    def test_is_complete(self):
        state = AlgoState(total_qty=100, filled_qty=100)
        assert state.is_complete

    def test_record_fill(self):
        from cpz.execution.algos.base import AlgoSlice

        s = AlgoSlice(seq=0, symbol="AAPL", side="buy", qty=50)
        state = AlgoState(total_qty=100, slices=[s])
        state.record_fill(0, 50, 150.0)
        assert state.filled_qty == 50
        assert state.avg_fill_price == 150.0
        assert s.status == "filled"

    def test_fill_pct(self):
        state = AlgoState(total_qty=200, filled_qty=50)
        assert state.fill_pct == 25.0



"""Tests for cpz.events — EventBus pub/sub with wildcards."""

from __future__ import annotations

import threading
from unittest.mock import MagicMock

from cpz.events.bus import EventBus, _topic_matches


class TestTopicMatching:
    def test_exact(self):
        assert _topic_matches("bar.AAPL.1Min", "bar.AAPL.1Min")

    def test_exact_mismatch(self):
        assert not _topic_matches("bar.AAPL.1Min", "bar.GOOG.1Min")

    def test_single_wildcard(self):
        assert _topic_matches("bar.AAPL.*", "bar.AAPL.1Min")
        assert _topic_matches("bar.AAPL.*", "bar.AAPL.1D")
        assert not _topic_matches("bar.AAPL.*", "bar.GOOG.1Min")

    def test_double_wildcard(self):
        assert _topic_matches("bar.**", "bar.AAPL.1Min")
        assert _topic_matches("bar.**", "bar.GOOG.1D")
        assert _topic_matches("order.**", "order.filled")
        assert not _topic_matches("bar.**", "quote.AAPL")

    def test_wildcard_middle(self):
        assert _topic_matches("bar.*.1Min", "bar.AAPL.1Min")
        assert not _topic_matches("bar.*.1Min", "bar.AAPL.1D")


class TestEventBus:
    def test_subscribe_publish(self):
        bus = EventBus()
        handler = MagicMock()
        bus.subscribe("test.topic", handler)
        bus.publish("test.topic", "data")
        handler.assert_called_once_with("data")

    def test_wildcard_subscribe(self):
        bus = EventBus()
        handler = MagicMock()
        bus.subscribe("bar.*.*", handler)
        bus.publish("bar.AAPL.1Min", "bar_data")
        handler.assert_called_once_with("bar_data")

    def test_double_wildcard(self):
        bus = EventBus()
        handler = MagicMock()
        bus.subscribe("order.**", handler)
        bus.publish("order.filled", "fill")
        bus.publish("order.rejected", "reject")
        assert handler.call_count == 2

    def test_unsubscribe(self):
        bus = EventBus()
        handler = MagicMock()
        bus.subscribe("test", handler)
        bus.unsubscribe("test", handler)
        bus.publish("test", "data")
        handler.assert_not_called()

    def test_no_duplicate_handlers(self):
        bus = EventBus()
        handler = MagicMock()
        bus.subscribe("test", handler)
        bus.subscribe("test", handler)
        bus.publish("test", "data")
        handler.assert_called_once()

    def test_has_subscribers(self):
        bus = EventBus()
        assert not bus.has_subscribers("test")
        handler = MagicMock()
        bus.subscribe("test", handler)
        assert bus.has_subscribers("test")

    def test_has_subscribers_wildcard(self):
        bus = EventBus()
        handler = MagicMock()
        bus.subscribe("bar.**", handler)
        assert bus.has_subscribers("bar.AAPL.1Min")

    def test_clear(self):
        bus = EventBus()
        handler = MagicMock()
        bus.subscribe("test", handler)
        bus.clear()
        assert not bus.has_subscribers("test")

    def test_publish_returns_count(self):
        bus = EventBus()
        h1, h2 = MagicMock(), MagicMock()
        bus.subscribe("test", h1)
        bus.subscribe("test", h2)
        count = bus.publish("test", "data")
        assert count == 2

    def test_handler_exception_doesnt_break_others(self):
        bus = EventBus()
        bad = MagicMock(side_effect=Exception("boom"))
        good = MagicMock()
        bus.subscribe("test", bad)
        bus.subscribe("test", good)
        bus.publish("test", "data")
        good.assert_called_once()

    def test_thread_safety(self):
        bus = EventBus()
        results = []

        def handler(data):
            results.append(data)

        bus.subscribe("test", handler)

        threads = [
            threading.Thread(target=bus.publish, args=("test", i))
            for i in range(100)
        ]
        for t in threads:
            t.start()
        for t in threads:
            t.join()

        assert len(results) == 100



"""Statistical indicators: z-score, rolling correlation, rolling beta, Hurst exponent, linear regression.

All functions operate on raw numpy arrays and return numpy arrays.
"""

from __future__ import annotations

from typing import Tuple

import numpy as np

from ._helpers import EPSILON


# ── Z-Score ──────────────────────────────────────────────────────────

def zscore_series(
    values: np.ndarray,
    period: int = 60,
    *,
    ddof: int = 1,
) -> np.ndarray:
    """Rolling z-score: ``(x - rolling_mean) / rolling_std``.

    Args:
        values: Input array.
        period: Rolling window size.
        ddof: Degrees of freedom for std (default 1 = sample std).
    """
    n = len(values)
    out = np.full(n, np.nan)
    for i in range(period - 1, n):
        window = values[i - period + 1 : i + 1]
        mean = np.mean(window)
        std = np.std(window, ddof=ddof)
        out[i] = (values[i] - mean) / std if std > EPSILON else 0.0
    return out


# ── Rolling Correlation ─────────────────────────────────────────────

def rolling_corr_series(
    series_a: np.ndarray,
    series_b: np.ndarray,
    period: int = 60,
) -> np.ndarray:
    """Rolling Pearson correlation between two series.

    Args:
        series_a, series_b: Input arrays (must be same length).
        period: Rolling window size.
    """
    n = min(len(series_a), len(series_b))
    out = np.full(n, np.nan)
    for i in range(period - 1, n):
        a = series_a[i - period + 1 : i + 1]
        b = series_b[i - period + 1 : i + 1]
        std_a = np.std(a, ddof=1)
        std_b = np.std(b, ddof=1)
        if std_a > EPSILON and std_b > EPSILON:
            out[i] = np.corrcoef(a, b)[0, 1]
        else:
            out[i] = 0.0
    return out


# ── Rolling Beta ─────────────────────────────────────────────────────

def rolling_beta_series(
    asset_returns: np.ndarray,
    benchmark_returns: np.ndarray,
    period: int = 60,
) -> np.ndarray:
    """Rolling OLS beta (slope of asset vs benchmark returns).

    Args:
        asset_returns: Asset return array.
        benchmark_returns: Benchmark return array.
        period: Rolling window size.
    """
    n = min(len(asset_returns), len(benchmark_returns))
    out = np.full(n, np.nan)
    for i in range(period - 1, n):
        y = asset_returns[i - period + 1 : i + 1]
        x = benchmark_returns[i - period + 1 : i + 1]
        var_x = np.var(x, ddof=1)
        if var_x > EPSILON:
            cov_xy = np.cov(x, y, ddof=1)[0, 1]
            out[i] = cov_xy / var_x
        else:
            out[i] = 0.0
    return out


# ── Hurst Exponent ───────────────────────────────────────────────────

def hurst_series(
    close: np.ndarray,
    max_lag: int = 100,
    *,
    min_lag: int = 2,
) -> float:
    """Hurst exponent via rescaled range (R/S) analysis.

    H < 0.5 → mean-reverting
    H = 0.5 → random walk
    H > 0.5 → trending

    Args:
        close: Price array.
        max_lag: Maximum lag for R/S computation.
        min_lag: Minimum lag (default 2).

    Returns:
        Scalar Hurst exponent.
    """
    log_prices = np.log(np.maximum(close, EPSILON))
    returns = np.diff(log_prices)
    n = len(returns)
    max_lag = min(max_lag, n // 2)

    if max_lag < min_lag:
        return 0.5

    lags = range(min_lag, max_lag + 1)
    rs_values = []
    lag_values = []

    for lag in lags:
        n_blocks = n // lag
        if n_blocks < 1:
            continue

        rs_sum = 0.0
        count = 0
        for b in range(n_blocks):
            block = returns[b * lag : (b + 1) * lag]
            mean_b = np.mean(block)
            deviations = np.cumsum(block - mean_b)
            r = np.max(deviations) - np.min(deviations)
            s = np.std(block, ddof=1)
            if s > EPSILON:
                rs_sum += r / s
                count += 1

        if count > 0:
            rs_values.append(np.log(rs_sum / count))
            lag_values.append(np.log(lag))

    if len(lag_values) < 2:
        return 0.5

    coeffs = np.polyfit(lag_values, rs_values, 1)
    return float(coeffs[0])


# ── Linear Regression ────────────────────────────────────────────────

def linear_reg_series(
    close: np.ndarray,
    period: int = 20,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """Rolling OLS linear regression on close prices.

    Args:
        close: Price array.
        period: Rolling window size.

    Returns:
        ``(slope, r_squared, intercept, forecast)`` arrays where
        *forecast* is the next-bar predicted value.
    """
    n = len(close)
    slope = np.full(n, np.nan)
    r_sq = np.full(n, np.nan)
    intercept = np.full(n, np.nan)
    forecast = np.full(n, np.nan)

    x = np.arange(period, dtype=np.float64)
    x_mean = np.mean(x)
    ss_x = np.sum((x - x_mean) ** 2)

    for i in range(period - 1, n):
        y = close[i - period + 1 : i + 1]
        y_mean = np.mean(y)
        ss_xy = np.sum((x - x_mean) * (y - y_mean))
        ss_y = np.sum((y - y_mean) ** 2)

        b = ss_xy / ss_x if ss_x > EPSILON else 0.0
        a = y_mean - b * x_mean
        slope[i] = b
        intercept[i] = a
        r_sq[i] = (ss_xy ** 2) / (ss_x * ss_y) if ss_y > EPSILON else 0.0
        forecast[i] = a + b * period

    return slope, r_sq, intercept, forecast



"""StrategyConfig — serializable, validated strategy configuration.

Subclass this to define your strategy's parameters. The same config works
for backtesting and live trading, and can be serialized for parameter sweeps,
version control, and remote execution.
"""

from __future__ import annotations

from typing import Any, Dict, List, Optional

from pydantic import BaseModel, Field


class StrategyConfig(BaseModel):
    """Base configuration for a Strategy.

    Subclass and add your own fields::

        class MyConfig(StrategyConfig):
            fast_period: int = 10
            slow_period: int = 30
            trade_size: float = 100

    The ``instruments`` list defines which symbols the strategy trades.
    ``strategy_id`` is auto-generated if not provided.
    """

    strategy_id: str = ""
    instruments: List[str] = Field(default_factory=list)
    broker: str = "alpaca"
    environment: str = "paper"
    parameters: Dict[str, Any] = Field(default_factory=dict)

    model_config = {"extra": "allow"}

    def to_dict(self) -> Dict[str, Any]:
        return self.model_dump()

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "StrategyConfig":
        return cls(**data)



"""CPZ Signals — strategy-level position sizing and portfolio risk filters.

Usage inside ``generate_signal()``::

    from cpz.signals import vol_target, max_positions, regime_filter

    def generate_signal(data, state):
        if not regime_filter(data.get("^VIX"), threshold=30):
            return {}
        weights = {}
        for sym in SYMBOLS:
            weights[sym] = vol_target(data[sym], risk=0.02)
        return max_positions(weights, limit=3)
"""

from .sizing import (
    atr_size,
    equal_weight,
    inverse_vol_weights,
    kelly,
    risk_parity_weights,
    vol_target,
)
from .filters import (
    long_only,
    max_drawdown_filter,
    max_exposure,
    max_positions,
    max_weight,
    min_weight,
    normalize,
    regime_filter,
    sector_limit,
    trend_filter,
)

__all__ = [
    # Sizing
    "vol_target", "atr_size", "kelly", "equal_weight",
    "risk_parity_weights", "inverse_vol_weights",
    # Filters
    "max_positions", "max_weight", "max_exposure", "normalize",
    "long_only", "sector_limit", "min_weight",
    "regime_filter", "trend_filter", "max_drawdown_filter",
]



"""CPZ Alpha — alpha research tools for signal quality assessment.

Implements the diagnostic toolkit from Grinold-Kahn's fundamental law:
``IR = IC x sqrt(Breadth) x TC``
"""

from .analysis import (
    ICResult,
    DecayResult,
    QuantileResult,
    TurnoverMetrics,
    information_coefficient,
    rolling_ic,
    signal_decay,
    quantile_returns,
    signal_turnover,
    breadth,
    transfer_coefficient,
)
from .combination import (
    FundLawResult,
    ic_weighted_blend,
    equal_weight_blend,
    zscore_blend,
    decay_adjusted_blend,
    fundamental_law_analysis,
)

__all__ = [
    # Analysis
    "information_coefficient", "rolling_ic", "signal_decay",
    "quantile_returns", "signal_turnover", "breadth", "transfer_coefficient",
    "ICResult", "DecayResult", "QuantileResult", "TurnoverMetrics",
    # Combination
    "ic_weighted_blend", "equal_weight_blend", "zscore_blend",
    "decay_adjusted_blend", "fundamental_law_analysis",
    "FundLawResult",
]



"""Typed domain model — validated, immutable value types for trading.

Inspired by NautilusTrader's fail-fast, precision-safe domain model.
All types reject NaN, Infinity, and out-of-range values at construction time.
"""

from __future__ import annotations

from .types import Price, Quantity, Money
from .events import (
    Event,
    BarEvent,
    QuoteEvent,
    TradeEvent,
    OrderEvent,
    FillEvent,
    PositionEvent,
    TimeEvent,
    SignalEvent,
)

__all__ = [
    "Price",
    "Quantity",
    "Money",
    "Event",
    "BarEvent",
    "QuoteEvent",
    "TradeEvent",
    "OrderEvent",
    "FillEvent",
    "PositionEvent",
    "TimeEvent",
    "SignalEvent",
]



"""Volume indicators: VWAP, OBV, CMF, volume z-score.

All functions operate on raw numpy arrays and return numpy arrays.
"""

from __future__ import annotations

import numpy as np

from ._helpers import EPSILON


# ── VWAP ─────────────────────────────────────────────────────────────

def vwap_series(
    high: np.ndarray,
    low: np.ndarray,
    close: np.ndarray,
    volume: np.ndarray,
    *,
    anchor_session: bool = True,
) -> np.ndarray:
    """Volume-Weighted Average Price.

    Args:
        high, low, close: Price arrays.
        volume: Volume array.
        anchor_session: If ``True`` (default), computes a cumulative VWAP
            across the full array.  Set ``False`` for a rolling variant
            (not standard but sometimes useful).
    """
    tp = (high + low + close) / 3.0
    cum_tp_vol = np.cumsum(tp * volume)
    cum_vol = np.cumsum(volume)
    with np.errstate(divide="ignore", invalid="ignore"):
        return np.where(cum_vol > EPSILON, cum_tp_vol / cum_vol, np.nan)


# ── OBV ──────────────────────────────────────────────────────────────

def obv_series(close: np.ndarray, volume: np.ndarray) -> np.ndarray:
    """On-Balance Volume."""
    n = len(close)
    out = np.zeros(n)
    for i in range(1, n):
        if close[i] > close[i - 1]:
            out[i] = out[i - 1] + volume[i]
        elif close[i] < close[i - 1]:
            out[i] = out[i - 1] - volume[i]
        else:
            out[i] = out[i - 1]
    return out


# ── CMF ──────────────────────────────────────────────────────────────

def cmf_series(
    high: np.ndarray,
    low: np.ndarray,
    close: np.ndarray,
    volume: np.ndarray,
    period: int = 20,
) -> np.ndarray:
    """Chaikin Money Flow.

    CMF = sum(MFV, period) / sum(volume, period)
    where MFV = ((close - low) - (high - close)) / (high - low) * volume
    """
    n = len(close)
    hl_range = high - low
    with np.errstate(divide="ignore", invalid="ignore"):
        mf_multiplier = np.where(
            hl_range > EPSILON,
            ((close - low) - (high - close)) / hl_range,
            0.0,
        )
    mfv = mf_multiplier * volume

    out = np.full(n, np.nan)
    for i in range(period - 1, n):
        vol_sum = np.sum(volume[i - period + 1 : i + 1])
        if vol_sum > EPSILON:
            out[i] = np.sum(mfv[i - period + 1 : i + 1]) / vol_sum
        else:
            out[i] = 0.0
    return out


# ── Volume Z-Score ───────────────────────────────────────────────────

def volume_zscore_series(
    volume: np.ndarray,
    period: int = 20,
) -> np.ndarray:
    """Volume z-score relative to its rolling mean and standard deviation.

    Args:
        volume: Volume array.
        period: Rolling look-back window for mean and std.
    """
    n = len(volume)
    out = np.full(n, np.nan)
    for i in range(period - 1, n):
        window = volume[i - period + 1 : i + 1]
        mean = np.mean(window)
        std = np.std(window, ddof=1)
        out[i] = (volume[i] - mean) / std if std > EPSILON else 0.0
    return out



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

from . import engine_pb2 as engine__pb2

GRPC_GENERATED_VERSION = '1.78.1'
GRPC_VERSION = grpc.__version__
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    raise RuntimeError(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + ' but the generated code in engine_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
    )


class EngineServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.PlaceOrder = channel.unary_unary(
                '/engine.EngineService/PlaceOrder',
                request_serializer=engine__pb2.OrderRequest.SerializeToString,
                response_deserializer=engine__pb2.OrderResponse.FromString,
                _registered_method=True)
        self.CancelOrder = channel.unary_unary(
                '/engine.EngineService/CancelOrder',
                request_serializer=engine__pb2.CancelRequest.SerializeToString,
                response_deserializer=engine__pb2.CancelResponse.FromString,
                _registered_method=True)
        self.GetPositions = channel.unary_unary(
                '/engine.EngineService/GetPositions',
                request_serializer=engine__pb2.PositionsRequest.SerializeToString,
                response_deserializer=engine__pb2.PositionsResponse.FromString,
                _registered_method=True)
        self.GetAccount = channel.unary_unary(
                '/engine.EngineService/GetAccount',
                request_serializer=engine__pb2.AccountRequest.SerializeToString,
                response_deserializer=engine__pb2.AccountResponse.FromString,
                _registered_method=True)
        self.GetOrder = channel.unary_unary(
                '/engine.EngineService/GetOrder',
                request_serializer=engine__pb2.GetOrderRequest.SerializeToString,
                response_deserializer=engine__pb2.OrderResponse.FromString,
                _registered_method=True)
        self.GetEngineStatus = channel.unary_unary(
                '/engine.EngineService/GetEngineStatus',
                request_serializer=engine__pb2.StatusRequest.SerializeToString,
                response_deserializer=engine__pb2.StatusResponse.FromString,
                _registered_method=True)
        self.ConfigureCredentials = channel.unary_unary(
                '/engine.EngineService/ConfigureCredentials',
                request_serializer=engine__pb2.CredentialsConfig.SerializeToString,
                response_deserializer=engine__pb2.ConfigResponse.FromString,
                _registered_method=True)
        self.UpdateConfig = channel.unary_unary(
                '/engine.EngineService/UpdateConfig',
                request_serializer=engine__pb2.ConfigUpdate.SerializeToString,
                response_deserializer=engine__pb2.ConfigResponse.FromString,
                _registered_method=True)
        self.StreamMarketData = channel.unary_stream(
                '/engine.EngineService/StreamMarketData',
                request_serializer=engine__pb2.SubscribeRequest.SerializeToString,
                response_deserializer=engine__pb2.MarketDataUpdate.FromString,
                _registered_method=True)
        self.StreamOrderUpdates = channel.unary_stream(
                '/engine.EngineService/StreamOrderUpdates',
                request_serializer=engine__pb2.OrderSubscribeRequest.SerializeToString,
                response_deserializer=engine__pb2.OrderUpdate.FromString,
                _registered_method=True)
        self.DeployStrategy = channel.unary_unary(
                '/engine.EngineService/DeployStrategy',
                request_serializer=engine__pb2.DeployStrategyRequest.SerializeToString,
                response_deserializer=engine__pb2.DeployStrategyResponse.FromString,
                _registered_method=True)
        self.StopStrategy = channel.unary_unary(
                '/engine.EngineService/StopStrategy',
                request_serializer=engine__pb2.StopStrategyRequest.SerializeToString,
                response_deserializer=engine__pb2.ConfigResponse.FromString,
                _registered_method=True)
        self.ListExecutors = channel.unary_unary(
                '/engine.EngineService/ListExecutors',
                request_serializer=engine__pb2.ListExecutorsRequest.SerializeToString,
                response_deserializer=engine__pb2.ListExecutorsResponse.FromString,
                _registered_method=True)
        self.GetExecutorStatus = channel.unary_unary(
                '/engine.EngineService/GetExecutorStatus',
                request_serializer=engine__pb2.GetExecutorStatusRequest.SerializeToString,
                response_deserializer=engine__pb2.ExecutorInfo.FromString,
                _registered_method=True)


class EngineServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def PlaceOrder(self, request, context):
        """--- Existing RPCs ---
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CancelOrder(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetPositions(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetAccount(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetOrder(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetEngineStatus(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ConfigureCredentials(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def UpdateConfig(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def StreamMarketData(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def StreamOrderUpdates(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DeployStrategy(self, request, context):
        """--- Multi-tenant executor management ---
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def StopStrategy(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListExecutors(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetExecutorStatus(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_EngineServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'PlaceOrder': grpc.unary_unary_rpc_method_handler(
                    servicer.PlaceOrder,
                    request_deserializer=engine__pb2.OrderRequest.FromString,
                    response_serializer=engine__pb2.OrderResponse.SerializeToString,
            ),
            'CancelOrder': grpc.unary_unary_rpc_method_handler(
                    servicer.CancelOrder,
                    request_deserializer=engine__pb2.CancelRequest.FromString,
                    response_serializer=engine__pb2.CancelResponse.SerializeToString,
            ),
            'GetPositions': grpc.unary_unary_rpc_method_handler(
                    servicer.GetPositions,
                    request_deserializer=engine__pb2.PositionsRequest.FromString,
                    response_serializer=engine__pb2.PositionsResponse.SerializeToString,
            ),
            'GetAccount': grpc.unary_unary_rpc_method_handler(
                    servicer.GetAccount,
                    request_deserializer=engine__pb2.AccountRequest.FromString,
                    response_serializer=engine__pb2.AccountResponse.SerializeToString,
            ),
            'GetOrder': grpc.unary_unary_rpc_method_handler(
                    servicer.GetOrder,
                    request_deserializer=engine__pb2.GetOrderRequest.FromString,
                    response_serializer=engine__pb2.OrderResponse.SerializeToString,
            ),
            'GetEngineStatus': grpc.unary_unary_rpc_method_handler(
                    servicer.GetEngineStatus,
                    request_deserializer=engine__pb2.StatusRequest.FromString,
                    response_serializer=engine__pb2.StatusResponse.SerializeToString,
            ),
            'ConfigureCredentials': grpc.unary_unary_rpc_method_handler(
                    servicer.ConfigureCredentials,
                    request_deserializer=engine__pb2.CredentialsConfig.FromString,
                    response_serializer=engine__pb2.ConfigResponse.SerializeToString,
            ),
            'UpdateConfig': grpc.unary_unary_rpc_method_handler(
                    servicer.UpdateConfig,
                    request_deserializer=engine__pb2.ConfigUpdate.FromString,
                    response_serializer=engine__pb2.ConfigResponse.SerializeToString,
            ),
            'StreamMarketData': grpc.unary_stream_rpc_method_handler(
                    servicer.StreamMarketData,
                    request_deserializer=engine__pb2.SubscribeRequest.FromString,
                    response_serializer=engine__pb2.MarketDataUpdate.SerializeToString,
            ),
            'StreamOrderUpdates': grpc.unary_stream_rpc_method_handler(
                    servicer.StreamOrderUpdates,
                    request_deserializer=engine__pb2.OrderSubscribeRequest.FromString,
                    response_serializer=engine__pb2.OrderUpdate.SerializeToString,
            ),
            'DeployStrategy': grpc.unary_unary_rpc_method_handler(
                    servicer.DeployStrategy,
                    request_deserializer=engine__pb2.DeployStrategyRequest.FromString,
                    response_serializer=engine__pb2.DeployStrategyResponse.SerializeToString,
            ),
            'StopStrategy': grpc.unary_unary_rpc_method_handler(
                    servicer.StopStrategy,
                    request_deserializer=engine__pb2.StopStrategyRequest.FromString,
                    response_serializer=engine__pb2.ConfigResponse.SerializeToString,
            ),
            'ListExecutors': grpc.unary_unary_rpc_method_handler(
                    servicer.ListExecutors,
                    request_deserializer=engine__pb2.ListExecutorsRequest.FromString,
                    response_serializer=engine__pb2.ListExecutorsResponse.SerializeToString,
            ),
            'GetExecutorStatus': grpc.unary_unary_rpc_method_handler(
                    servicer.GetExecutorStatus,
                    request_deserializer=engine__pb2.GetExecutorStatusRequest.FromString,
                    response_serializer=engine__pb2.ExecutorInfo.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'engine.EngineService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('engine.EngineService', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class EngineService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def PlaceOrder(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/engine.EngineService/PlaceOrder',
            engine__pb2.OrderRequest.SerializeToString,
            engine__pb2.OrderResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def CancelOrder(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/engine.EngineService/CancelOrder',
            engine__pb2.CancelRequest.SerializeToString,
            engine__pb2.CancelResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetPositions(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/engine.EngineService/GetPositions',
            engine__pb2.PositionsRequest.SerializeToString,
            engine__pb2.PositionsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetAccount(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/engine.EngineService/GetAccount',
            engine__pb2.AccountRequest.SerializeToString,
            engine__pb2.AccountResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetOrder(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/engine.EngineService/GetOrder',
            engine__pb2.GetOrderRequest.SerializeToString,
            engine__pb2.OrderResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetEngineStatus(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/engine.EngineService/GetEngineStatus',
            engine__pb2.StatusRequest.SerializeToString,
            engine__pb2.StatusResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ConfigureCredentials(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/engine.EngineService/ConfigureCredentials',
            engine__pb2.CredentialsConfig.SerializeToString,
            engine__pb2.ConfigResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def UpdateConfig(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/engine.EngineService/UpdateConfig',
            engine__pb2.ConfigUpdate.SerializeToString,
            engine__pb2.ConfigResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def StreamMarketData(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(
            request,
            target,
            '/engine.EngineService/StreamMarketData',
            engine__pb2.SubscribeRequest.SerializeToString,
            engine__pb2.MarketDataUpdate.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def StreamOrderUpdates(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(
            request,
            target,
            '/engine.EngineService/StreamOrderUpdates',
            engine__pb2.OrderSubscribeRequest.SerializeToString,
            engine__pb2.OrderUpdate.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def DeployStrategy(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/engine.EngineService/DeployStrategy',
            engine__pb2.DeployStrategyRequest.SerializeToString,
            engine__pb2.DeployStrategyResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def StopStrategy(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/engine.EngineService/StopStrategy',
            engine__pb2.StopStrategyRequest.SerializeToString,
            engine__pb2.ConfigResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ListExecutors(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/engine.EngineService/ListExecutors',
            engine__pb2.ListExecutorsRequest.SerializeToString,
            engine__pb2.ListExecutorsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetExecutorStatus(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/engine.EngineService/GetExecutorStatus',
            engine__pb2.GetExecutorStatusRequest.SerializeToString,
            engine__pb2.ExecutorInfo.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)



"""Precision-safe value types for trading: Price, Quantity, Money.

These types enforce data integrity at the boundary — rejecting NaN, Infinity,
negative quantities, and arithmetic that would silently overflow. Backed by
Python's Decimal for exact representation of financial values.
"""

from __future__ import annotations

import math
from decimal import Decimal, InvalidOperation, ROUND_HALF_UP
from typing import Union

_Numeric = Union[int, float, str, Decimal]


def _to_decimal(value: _Numeric, name: str) -> Decimal:
    """Convert a numeric input to Decimal, rejecting NaN and Infinity."""
    if isinstance(value, float):
        if math.isnan(value) or math.isinf(value):
            raise ValueError(f"{name} must be finite, got {value}")
        value = Decimal(str(value))
    elif isinstance(value, (int, str)):
        try:
            value = Decimal(str(value))
        except InvalidOperation as exc:
            raise ValueError(f"{name}: invalid numeric value {value!r}") from exc
    elif not isinstance(value, Decimal):
        raise TypeError(f"{name} must be numeric, got {type(value).__name__}")

    if value.is_nan() or value.is_infinite():
        raise ValueError(f"{name} must be finite, got {value}")
    return value


class Price:
    """Immutable, precision-safe price value.

    Rejects NaN, Infinity at construction. Supports arithmetic with other
    Price/Quantity/Money types with type-safe returns.
    """

    __slots__ = ("_value", "_precision")

    def __init__(self, value: _Numeric, precision: int | None = None) -> None:
        self._value = _to_decimal(value, "Price")
        if precision is not None:
            if precision < 0 or precision > 18:
                raise ValueError(f"precision must be 0-18, got {precision}")
            quant = Decimal(10) ** -precision
            self._value = self._value.quantize(quant, rounding=ROUND_HALF_UP)
            self._precision = precision
        else:
            self._precision = max(0, -self._value.as_tuple().exponent)

    @classmethod
    def from_str(cls, s: str, precision: int | None = None) -> "Price":
        return cls(s, precision=precision)

    @classmethod
    def from_float(cls, f: float, precision: int = 2) -> "Price":
        return cls(f, precision=precision)

    @property
    def value(self) -> Decimal:
        return self._value

    @property
    def precision(self) -> int:
        return self._precision

    def as_float(self) -> float:
        return float(self._value)

    def __repr__(self) -> str:
        return f"Price('{self._value}')"

    def __str__(self) -> str:
        return str(self._value)

    def __float__(self) -> float:
        return float(self._value)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Price):
            return self._value == other._value
        if isinstance(other, (int, float, Decimal)):
            return self._value == _to_decimal(other, "compare")
        return NotImplemented

    def __lt__(self, other: object) -> bool:
        if isinstance(other, Price):
            return self._value < other._value
        if isinstance(other, (int, float, Decimal)):
            return self._value < _to_decimal(other, "compare")
        return NotImplemented

    def __le__(self, other: object) -> bool:
        if isinstance(other, Price):
            return self._value <= other._value
        if isinstance(other, (int, float, Decimal)):
            return self._value <= _to_decimal(other, "compare")
        return NotImplemented

    def __gt__(self, other: object) -> bool:
        if isinstance(other, Price):
            return self._value > other._value
        if isinstance(other, (int, float, Decimal)):
            return self._value > _to_decimal(other, "compare")
        return NotImplemented

    def __ge__(self, other: object) -> bool:
        if isinstance(other, Price):
            return self._value >= other._value
        if isinstance(other, (int, float, Decimal)):
            return self._value >= _to_decimal(other, "compare")
        return NotImplemented

    def __hash__(self) -> int:
        return hash(self._value)

    def __add__(self, other: object) -> "Price":
        if isinstance(other, Price):
            return Price(self._value + other._value)
        if isinstance(other, (int, float, Decimal)):
            return Price(self._value + _to_decimal(other, "add"))
        return NotImplemented

    def __sub__(self, other: object) -> "Price":
        if isinstance(other, Price):
            return Price(self._value - other._value)
        if isinstance(other, (int, float, Decimal)):
            return Price(self._value - _to_decimal(other, "sub"))
        return NotImplemented

    def __mul__(self, other: object) -> "Money":
        if isinstance(other, Quantity):
            return Money(self._value * other._value)
        if isinstance(other, (int, float, Decimal)):
            return Money(self._value * _to_decimal(other, "mul"))
        return NotImplemented

    def __neg__(self) -> "Price":
        return Price(-self._value)

    def __abs__(self) -> "Price":
        return Price(abs(self._value))

    def __bool__(self) -> bool:
        return self._value != 0


class Quantity:
    """Immutable, non-negative quantity value.

    Rejects NaN, Infinity, and negative values at construction.
    """

    __slots__ = ("_value", "_precision")

    def __init__(self, value: _Numeric, precision: int | None = None) -> None:
        self._value = _to_decimal(value, "Quantity")
        if self._value < 0:
            raise ValueError(f"Quantity cannot be negative, got {self._value}")
        if precision is not None:
            if precision < 0 or precision > 18:
                raise ValueError(f"precision must be 0-18, got {precision}")
            quant = Decimal(10) ** -precision
            self._value = self._value.quantize(quant, rounding=ROUND_HALF_UP)
            self._precision = precision
        else:
            self._precision = max(0, -self._value.as_tuple().exponent)

    @classmethod
    def from_int(cls, n: int) -> "Quantity":
        return cls(n, precision=0)

    @classmethod
    def from_float(cls, f: float, precision: int = 2) -> "Quantity":
        return cls(f, precision=precision)

    @property
    def value(self) -> Decimal:
        return self._value

    @property
    def precision(self) -> int:
        return self._precision

    def as_float(self) -> float:
        return float(self._value)

    def __repr__(self) -> str:
        return f"Quantity('{self._value}')"

    def __str__(self) -> str:
        return str(self._value)

    def __float__(self) -> float:
        return float(self._value)

    def __int__(self) -> int:
        return int(self._value)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Quantity):
            return self._value == other._value
        if isinstance(other, (int, float, Decimal)):
            return self._value == _to_decimal(other, "compare")
        return NotImplemented

    def __lt__(self, other: object) -> bool:
        if isinstance(other, Quantity):
            return self._value < other._value
        if isinstance(other, (int, float, Decimal)):
            return self._value < _to_decimal(other, "compare")
        return NotImplemented

    def __le__(self, other: object) -> bool:
        if isinstance(other, Quantity):
            return self._value <= other._value
        if isinstance(other, (int, float, Decimal)):
            return self._value <= _to_decimal(other, "compare")
        return NotImplemented

    def __gt__(self, other: object) -> bool:
        if isinstance(other, Quantity):
            return self._value > other._value
        if isinstance(other, (int, float, Decimal)):
            return self._value > _to_decimal(other, "compare")
        return NotImplemented

    def __ge__(self, other: object) -> bool:
        if isinstance(other, Quantity):
            return self._value >= other._value
        if isinstance(other, (int, float, Decimal)):
            return self._value >= _to_decimal(other, "compare")
        return NotImplemented

    def __hash__(self) -> int:
        return hash(self._value)

    def __add__(self, other: object) -> "Quantity":
        if isinstance(other, Quantity):
            return Quantity(self._value + other._value)
        if isinstance(other, (int, float, Decimal)):
            return Quantity(self._value + _to_decimal(other, "add"))
        return NotImplemented

    def __sub__(self, other: object) -> "Quantity":
        if isinstance(other, Quantity):
            return Quantity(self._value - other._value)
        if isinstance(other, (int, float, Decimal)):
            return Quantity(self._value - _to_decimal(other, "sub"))
        return NotImplemented

    def __mul__(self, other: object) -> "Quantity":
        if isinstance(other, (int, float, Decimal)):
            return Quantity(self._value * _to_decimal(other, "mul"))
        return NotImplemented

    def __bool__(self) -> bool:
        return self._value != 0


class Money:
    """Immutable monetary value with currency tracking.

    Supports both positive and negative amounts (for P&L, debits, etc.).
    """

    __slots__ = ("_value", "_currency")

    def __init__(self, value: _Numeric, currency: str = "USD") -> None:
        self._value = _to_decimal(value, "Money")
        self._currency = currency.upper()

    @classmethod
    def zero(cls, currency: str = "USD") -> "Money":
        return cls(0, currency)

    @property
    def value(self) -> Decimal:
        return self._value

    @property
    def currency(self) -> str:
        return self._currency

    def as_float(self) -> float:
        return float(self._value)

    def __repr__(self) -> str:
        return f"Money('{self._value}', '{self._currency}')"

    def __str__(self) -> str:
        return f"{self._value} {self._currency}"

    def __float__(self) -> float:
        return float(self._value)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Money):
            return self._value == other._value and self._currency == other._currency
        if isinstance(other, (int, float, Decimal)):
            return self._value == _to_decimal(other, "compare")
        return NotImplemented

    def __lt__(self, other: object) -> bool:
        if isinstance(other, Money):
            if self._currency != other._currency:
                raise ValueError(f"Cannot compare {self._currency} with {other._currency}")
            return self._value < other._value
        if isinstance(other, (int, float, Decimal)):
            return self._value < _to_decimal(other, "compare")
        return NotImplemented

    def __le__(self, other: object) -> bool:
        if isinstance(other, Money):
            if self._currency != other._currency:
                raise ValueError(f"Cannot compare {self._currency} with {other._currency}")
            return self._value <= other._value
        if isinstance(other, (int, float, Decimal)):
            return self._value <= _to_decimal(other, "compare")
        return NotImplemented

    def __gt__(self, other: object) -> bool:
        if isinstance(other, Money):
            if self._currency != other._currency:
                raise ValueError(f"Cannot compare {self._currency} with {other._currency}")
            return self._value > other._value
        if isinstance(other, (int, float, Decimal)):
            return self._value > _to_decimal(other, "compare")
        return NotImplemented

    def __ge__(self, other: object) -> bool:
        if isinstance(other, Money):
            if self._currency != other._currency:
                raise ValueError(f"Cannot compare {self._currency} with {other._currency}")
            return self._value >= other._value
        if isinstance(other, (int, float, Decimal)):
            return self._value >= _to_decimal(other, "compare")
        return NotImplemented

    def __hash__(self) -> int:
        return hash((self._value, self._currency))

    def __add__(self, other: object) -> "Money":
        if isinstance(other, Money):
            if self._currency != other._currency:
                raise ValueError(f"Cannot add {self._currency} and {other._currency}")
            return Money(self._value + other._value, self._currency)
        if isinstance(other, (int, float, Decimal)):
            return Money(self._value + _to_decimal(other, "add"), self._currency)
        return NotImplemented

    def __sub__(self, other: object) -> "Money":
        if isinstance(other, Money):
            if self._currency != other._currency:
                raise ValueError(f"Cannot subtract {self._currency} and {other._currency}")
            return Money(self._value - other._value, self._currency)
        if isinstance(other, (int, float, Decimal)):
            return Money(self._value - _to_decimal(other, "sub"), self._currency)
        return NotImplemented

    def __mul__(self, other: object) -> "Money":
        if isinstance(other, (int, float, Decimal)):
            return Money(self._value * _to_decimal(other, "mul"), self._currency)
        return NotImplemented

    def __neg__(self) -> "Money":
        return Money(-self._value, self._currency)

    def __abs__(self) -> "Money":
        return Money(abs(self._value), self._currency)

    def __bool__(self) -> bool:
        return self._value != 0



"""BacktestEngine — deterministic historical replay with fill simulation.

Replays bars through strategy handlers in timestamp order, simulates
fills with configurable slippage and commission models, and produces
a ``BacktestResult`` with equity curve, trade log, and full analytics.
"""

from __future__ import annotations

import logging
import uuid
from dataclasses import dataclass, field
from datetime import datetime
from typing import TYPE_CHECKING, Any, Dict, List, Optional

import numpy as np

from ..events.bus import EventBus
from ..execution.enums import OrderSide, OrderType
from ..execution.models import Order, OrderSubmitRequest
from ..model.events import BarEvent, FillEvent, OrderEvent
from .base import Strategy
from .context import Cache, Clock, Portfolio

if TYPE_CHECKING:
    from ..clients.sync import CPZClient

logger = logging.getLogger(__name__)


@dataclass
class TradeRecord:
    """Single round-trip or open trade captured during backtest."""

    symbol: str
    side: str
    qty: float
    entry_price: float
    exit_price: float = 0.0
    entry_time: str = ""
    exit_time: str = ""
    pnl: float = 0.0
    commission: float = 0.0
    strategy_id: str = ""


@dataclass
class BacktestResult:
    """Complete results from a backtest run."""

    equity_curve: List[float] = field(default_factory=list)
    timestamps: List[str] = field(default_factory=list)
    trades: List[TradeRecord] = field(default_factory=list)
    initial_cash: float = 100_000.0
    final_equity: float = 0.0
    total_return_pct: float = 0.0
    sharpe_ratio: float = 0.0
    sortino_ratio: float = 0.0
    max_drawdown_pct: float = 0.0
    win_rate: float = 0.0
    profit_factor: float = 0.0
    total_trades: int = 0
    winning_trades: int = 0
    losing_trades: int = 0
    avg_win: float = 0.0
    avg_loss: float = 0.0
    total_commission: float = 0.0
    calmar_ratio: float = 0.0
    avg_trade_pnl: float = 0.0
    max_consecutive_wins: int = 0
    max_consecutive_losses: int = 0
    start_date: str = ""
    end_date: str = ""
    timeframe: str = ""

    def summary(self) -> str:
        """Human-readable summary of backtest results."""
        lines = [
            "=" * 60,
            "  BACKTEST RESULTS",
            "=" * 60,
            f"  Period:            {self.start_date} → {self.end_date}",
            f"  Timeframe:         {self.timeframe}",
            f"  Initial Cash:      ${self.initial_cash:,.2f}",
            f"  Final Equity:      ${self.final_equity:,.2f}",
            f"  Total Return:      {self.total_return_pct:+.2f}%",
            "-" * 60,
            f"  Sharpe Ratio:      {self.sharpe_ratio:.3f}",
            f"  Sortino Ratio:     {self.sortino_ratio:.3f}",
            f"  Calmar Ratio:      {self.calmar_ratio:.3f}",
            f"  Max Drawdown:      {self.max_drawdown_pct:.2f}%",
            "-" * 60,
            f"  Total Trades:      {self.total_trades}",
            f"  Win Rate:          {self.win_rate:.1f}%",
            f"  Profit Factor:     {self.profit_factor:.2f}",
            f"  Avg Win:           ${self.avg_win:,.2f}",
            f"  Avg Loss:          ${self.avg_loss:,.2f}",
            f"  Avg Trade P&L:     ${self.avg_trade_pnl:,.2f}",
            f"  Commission:        ${self.total_commission:,.2f}",
            f"  Max Consec Wins:   {self.max_consecutive_wins}",
            f"  Max Consec Losses: {self.max_consecutive_losses}",
            "=" * 60,
        ]
        return "\n".join(lines)


class _SimulatedRouter:
    """Minimal router substitute for the backtest engine.

    Simulates order fills at the next bar's open with slippage.
    """

    def __init__(
        self,
        portfolio: Portfolio,
        bus: EventBus,
        slippage_pct: float = 0.0001,
        commission_pct: float = 0.0003,
    ) -> None:
        self._portfolio = portfolio
        self._bus = bus
        self._slippage_pct = slippage_pct
        self._commission_pct = commission_pct
        self._pending: List[OrderSubmitRequest] = []
        self._filled_orders: List[TradeRecord] = []
        self._current_bar: Optional[BarEvent] = None

    def submit_order(self, req: OrderSubmitRequest) -> Order:
        oid = uuid.uuid4().hex[:12]
        self._pending.append(req)
        return Order(
            id=oid,
            symbol=req.symbol,
            side=req.side,
            qty=req.qty or 0.0,
            type=req.type,
            time_in_force=req.time_in_force,
            status="pending",
        )

    def cancel_order(self, order_id: str) -> Order:
        return Order(
            id=order_id, symbol="", side=OrderSide.BUY, qty=0,
            type=OrderType.MARKET, time_in_force="DAY", status="canceled",
        )

    def list_orders(self, status: str | None = None, **kw: Any) -> list:
        return []

    def process_pending(self, bar: BarEvent, strategies: List[Strategy]) -> None:
        """Fill pending orders at the bar's open price with slippage."""
        self._current_bar = bar
        pending = [r for r in self._pending if r.symbol == bar.symbol]
        self._pending = [r for r in self._pending if r.symbol != bar.symbol]

        for req in pending:
            fill_price = bar.open
            if req.side == OrderSide.BUY:
                fill_price *= 1 + self._slippage_pct
            else:
                fill_price *= 1 - self._slippage_pct

            qty = req.qty or (req.notional / fill_price if req.notional else 0)
            if qty <= 0:
                continue

            commission = qty * fill_price * self._commission_pct

            fill = FillEvent(
                order_id=uuid.uuid4().hex[:12],
                symbol=req.symbol,
                side=req.side.value,
                fill_qty=qty,
                fill_price=fill_price,
                commission=commission,
                strategy_id=req.strategy_id,
            )

            for strat in strategies:
                if strat.strategy_id == req.strategy_id:
                    strat._handle_fill(fill)
                    break

            ts = datetime.utcfromtimestamp(bar.ts_event / 1e9).isoformat() if bar.ts_event else ""
            self._filled_orders.append(
                TradeRecord(
                    symbol=req.symbol,
                    side=req.side.value,
                    qty=qty,
                    entry_price=fill_price,
                    entry_time=ts,
                    commission=commission,
                    strategy_id=req.strategy_id,
                )
            )


class BacktestEngine:
    """Historical replay engine with deterministic fill simulation."""

    def __init__(
        self,
        client: "CPZClient",
        bus: EventBus,
        initial_cash: float = 100_000.0,
        commission_pct: float = 0.0003,
        slippage_pct: float = 0.0001,
    ) -> None:
        self._client = client
        self._bus = bus
        self._initial_cash = initial_cash
        self._commission_pct = commission_pct
        self._slippage_pct = slippage_pct

    def run(
        self,
        strategies: List[Strategy],
        start: str,
        end: str,
        timeframe: str = "1D",
    ) -> BacktestResult:
        portfolio = Portfolio(self._initial_cash)
        sim_router = _SimulatedRouter(
            portfolio=portfolio,
            bus=self._bus,
            slippage_pct=self._slippage_pct,
            commission_pct=self._commission_pct,
        )
        clock = Clock()

        all_symbols: set[str] = set()
        for strat in strategies:
            strat.portfolio = portfolio
            strat.clock = clock
            strat.cache = Cache()
            strat._router = sim_router  # type: ignore[assignment]
            strat._bus = self._bus
            for inst in strat.config.instruments:
                all_symbols.add(inst)

        for strat in strategies:
            strat._running = True
            strat.on_start()
            for sub in strat._subscriptions:
                all_symbols.add(sub["symbol"])

        if not all_symbols:
            raise ValueError("No instruments to backtest — set config.instruments or subscribe_bars()")

        symbol_bars = self._fetch_data(list(all_symbols), start, end, timeframe)

        merged = []
        for symbol, bars_list in symbol_bars.items():
            for bar in bars_list:
                merged.append((bar, symbol))

        merged.sort(key=lambda x: x[0].ts_event)

        equity_curve: List[float] = [self._initial_cash]
        timestamps: List[str] = [start]

        for bar, symbol in merged:
            clock.set_time(bar.ts_event)
            sim_router.process_pending(bar, strategies)
            portfolio.update_price(symbol, bar.close)

            topic = f"bar.{symbol}.{timeframe}"
            self._bus.publish(topic, bar)

            for strat in strategies:
                has_sub = any(
                    s["symbol"] == symbol and s.get("timeframe", "1D") == timeframe
                    for s in strat._subscriptions
                )
                if has_sub:
                    strat._handle_bar(bar)

            equity_curve.append(portfolio.equity())
            ts_str = datetime.utcfromtimestamp(bar.ts_event / 1e9).isoformat() if bar.ts_event else ""
            timestamps.append(ts_str)

        for strat in strategies:
            strat.on_stop()
            strat._running = False

        return self._build_result(
            equity_curve=equity_curve,
            timestamps=timestamps,
            trades=sim_router._filled_orders,
            portfolio=portfolio,
            start=start,
            end=end,
            timeframe=timeframe,
        )

    def _fetch_data(
        self, symbols: List[str], start: str, end: str, timeframe: str
    ) -> Dict[str, List[BarEvent]]:
        """Fetch historical bars via the DataClient and convert to BarEvents."""
        result: Dict[str, List[BarEvent]] = {}

        for symbol in symbols:
            try:
                df = self._client.data.bars(
                    symbol, timeframe=timeframe, start=start, end=end, limit=10000
                )
            except Exception as exc:
                logger.warning("Failed to fetch data for %s: %s", symbol, exc)
                continue

            bars: List[BarEvent] = []
            if df is None:
                continue

            try:
                import pandas as pd

                if isinstance(df, pd.DataFrame) and not df.empty:
                    for idx, row in df.iterrows():
                        ts = idx if isinstance(idx, pd.Timestamp) else pd.Timestamp(str(idx))
                        ts_ns = int(ts.timestamp() * 1e9)
                        bars.append(
                            BarEvent(
                                symbol=symbol,
                                timeframe=timeframe,
                                open=float(row.get("open", row.get("o", 0))),
                                high=float(row.get("high", row.get("h", 0))),
                                low=float(row.get("low", row.get("l", 0))),
                                close=float(row.get("close", row.get("c", 0))),
                                volume=float(row.get("volume", row.get("v", 0))),
                                vwap=float(row.get("vwap", row.get("vw", 0))),
                                ts_event=ts_ns,
                                ts_init=ts_ns,
                            )
                        )
            except ImportError:
                pass

            if bars:
                result[symbol] = bars
                logger.info("Loaded %d bars for %s", len(bars), symbol)

        return result

    def _build_result(
        self,
        equity_curve: List[float],
        timestamps: List[str],
        trades: List[TradeRecord],
        portfolio: Portfolio,
        start: str,
        end: str,
        timeframe: str,
    ) -> BacktestResult:
        final_eq = equity_curve[-1] if equity_curve else self._initial_cash
        total_ret = ((final_eq / self._initial_cash) - 1) * 100

        daily_returns: List[float] = []
        for i in range(1, len(equity_curve)):
            if equity_curve[i - 1] != 0:
                daily_returns.append(equity_curve[i] / equity_curve[i - 1] - 1)

        sharpe = sortino = max_dd = calmar = 0.0
        if daily_returns:
            arr = np.array(daily_returns)
            mean_r = float(np.mean(arr))
            std_r = float(np.std(arr, ddof=1)) if len(arr) > 1 else 0.0
            sharpe = (mean_r / std_r * np.sqrt(252)) if std_r > 0 else 0.0

            downside = arr[arr < 0]
            down_std = float(np.std(downside, ddof=1)) if len(downside) > 1 else 0.0
            sortino = (mean_r / down_std * np.sqrt(252)) if down_std > 0 else 0.0

            cumulative = np.cumprod(1 + arr)
            running_max = np.maximum.accumulate(cumulative)
            drawdowns = (cumulative - running_max) / running_max
            max_dd = float(np.min(drawdowns)) * 100
            calmar = (total_ret / abs(max_dd)) if max_dd != 0 else 0.0

        closed_pnls = [t.pnl for t in trades if t.exit_price > 0]
        total_t = len(trades)
        wins = [p for p in closed_pnls if p > 0]
        losses = [p for p in closed_pnls if p <= 0]
        win_rate = (len(wins) / len(closed_pnls) * 100) if closed_pnls else 0.0
        avg_win = float(np.mean(wins)) if wins else 0.0
        avg_loss = float(np.mean(losses)) if losses else 0.0
        gross_profit = sum(wins) if wins else 0.0
        gross_loss = abs(sum(losses)) if losses else 0.0
        profit_factor = (gross_profit / gross_loss) if gross_loss > 0 else float("inf")
        avg_pnl = float(np.mean(closed_pnls)) if closed_pnls else 0.0

        max_cw = max_cl = cw = cl = 0
        for p in closed_pnls:
            if p > 0:
                cw += 1
                cl = 0
            else:
                cl += 1
                cw = 0
            max_cw = max(max_cw, cw)
            max_cl = max(max_cl, cl)

        return BacktestResult(
            equity_curve=equity_curve,
            timestamps=timestamps,
            trades=trades,
            initial_cash=self._initial_cash,
            final_equity=final_eq,
            total_return_pct=total_ret,
            sharpe_ratio=sharpe,
            sortino_ratio=sortino,
            max_drawdown_pct=max_dd,
            win_rate=win_rate,
            profit_factor=profit_factor,
            total_trades=total_t,
            winning_trades=len(wins),
            losing_trades=len(losses),
            avg_win=avg_win,
            avg_loss=avg_loss,
            total_commission=portfolio.commission_total,
            calmar_ratio=calmar,
            avg_trade_pnl=avg_pnl,
            max_consecutive_wins=max_cw,
            max_consecutive_losses=max_cl,
            start_date=start,
            end_date=end,
            timeframe=timeframe,
        )
